#!/usr/bin/env python3
#
#  formula.py
"""
Parse formulae into a Python object.
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on ChemPy (https://github.com/bjodah/chempy)
#  |  Copyright (c) 2015-2018, Björn Dahlgren
#  |  All rights reserved.
#  |
#  |  Redistribution and use in source and binary forms, with or without modification,
#  |  are permitted provided that the following conditions are met:
#  |
#  |    Redistributions of source code must retain the above copyright notice, this
#  |    list of conditions and the following disclaimer.
#  |
#  |    Redistributions in binary form must reproduce the above copyright notice, this
#  |    list of conditions and the following disclaimer in the documentation and/or
#  |    other materials provided with the distribution.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  |  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  |  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  |  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
#  |  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  |  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  |  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#  |  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  |  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  |  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#  Also based on Pyteomics (https://github.com/levitsky/pyteomics)
#  |  Copyright (c) 2011-2015, Anton Goloborodko & Lev Levitsky
#  |  Licensed under the Apache License, Version 2.0 (the "License");
#  |  you may not use this file except in compliance with the License.
#  |  You may obtain a copy of the License at
#  |
#  |    http://www.apache.org/licenses/LICENSE-2.0
#  |
#  |  Unless required by applicable law or agreed to in writing, software
#  |  distributed under the License is distributed on an "AS IS" BASIS,
#  |  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  |  See the License for the specific language governing permissions and
#  |  limitations under the License.
#  |
#  |  See also:
#  |  Goloborodko, A.A.; Levitsky, L.I.; Ivanov, M.V.; and Gorshkov, M.V. (2013)
#  |  "Pyteomics - a Python Framework for Exploratory Data Analysis and Rapid Software
#  |  Prototyping in Proteomics", Journal of The American Society for Mass Spectrometry,
#  |  24(2), 301–304. DOI: `10.1007/s13361-012-0516-6 <http://dx.doi.org/10.1007/s13361-012-0516-6>`_
#  |
#  |  Levitsky, L.I.; Klein, J.; Ivanov, M.V.; and Gorshkov, M.V. (2018)
#  |  "Pyteomics 4.0: five years of development of a Python proteomics framework",
#  |  Journal of Proteome Research.
#  |  DOI: `10.1021/acs.jproteome.8b00717 <http://dx.doi.org/10.1021/acs.jproteome.8b00717>`_
#
#  Also based on molmass (https://github.com/cgohlke/molmass)
#  |  Copyright (c) 1990-2020, Christoph Gohlke
#  |  All rights reserved.
#  |  Licensed under the BSD 3-Clause License
#  |  Redistribution and use in source and binary forms, with or without
#  |  modification, are permitted provided that the following conditions are met:
#  |
#  |  1. Redistributions of source code must retain the above copyright notice,
#  |     this list of conditions and the following disclaimer.
#  |
#  |  2. Redistributions in binary form must reproduce the above copyright notice,
#  |     this list of conditions and the following disclaimer in the documentation
#  |     and/or other materials provided with the distribution.
#  |
#  |  3. Neither the name of the copyright holder nor the names of its
#  |     contributors may be used to endorse or promote products derived from
#  |     this software without specific prior written permission.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  |  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  |  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  |  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  |  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  |  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  |  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  |  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  |  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  |  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  |  POSSIBILITY OF SUCH DAMAGE.
#  |

# stdlib
import math
from collections import Counter, defaultdict
from itertools import combinations_with_replacement, product
from typing import Dict, Iterator, List, Optional, Sequence, Tuple, Type, TypeVar

# 3rd party
from domdf_python_tools.doctools import prettify_docstrings
from mathematical.utils import gcd_array  # nodep

# this package
from chemistry_tools.elements import ELEMENTS, D, T, isotope_data
from chemistry_tools.formulae.parser import string_to_composition

# this package
from ._parser_core import _make_isotope_string
from .composition import Composition
from .iso_dist import IsotopeDistribution
from .utils import GROUPS, hill_order, split_isotope

__all__ = ["Formula", 'F']

F = TypeVar('F', bound="Formula")


@prettify_docstrings
class Formula(defaultdict, Counter):
	"""
	A Formula object stores a chemical composition of a compound.
	It is based on :class:`dict`, with the symbols of chemical elements
	as keys and the values equal to the number of atoms of the corresponding
	element in the compound.

	:param composition: A :class:`~chemistry_tools.formulae.formula.Formula` object with the elemental
		composition of a substance, or a :class:`python:dict` representing the same.
		If :py:obj:`None` an empty object is created
	:param charge:

	.. autosummary-widths:: 55/100
	"""

	# TODO: option to convert D and T to H[2] and H[3] ("heavy hydrogen")

	def __init__(self, composition: Optional[Dict[str, int]] = None, charge: int = 0):
		defaultdict.__init__(self, int)

		self.charge = 0

		if composition is not None:
			for isotope_string, num_atoms in composition.items():
				element_name, isotope_num = split_isotope(isotope_string)

				# Remove explicitly undefined isotopes (e.g. X[0]).
				self[_make_isotope_string(element_name, isotope_num)] = num_atoms

			if isinstance(composition, Formula):
				if composition.charge:
					if charge and composition.charge:
						if charge != composition.charge:
							# No point raising error if the charges are the same
							raise ValueError(
									"Charge is specified in both the composition and with the "
									f"'charge' keyword argument. {charge}, {self.charge}"
									)

					charge = composition.charge

		self._set_charge(charge)

	def _set_charge(self, charge: int) -> None:
		# Get charge
		if charge and self.charge:
			if charge != self.charge:
				# No point raising error if the charges are the same
				raise ValueError(
						"Charge is specified in both the formula and with the "
						f"'charge' keyword argument. {charge}, {self.charge}"
						)
		elif charge:
			self.charge = charge

	@classmethod
	def from_string(cls: Type['F'], formula: str, charge: int = 0) -> F:
		"""
		Create a new :class:`~chemistry_tools.formulae.formula.Formula` object by parsing a string.

		.. latex:vspace:: -5px

		.. note:: Isotopes cannot (currently) be parsed using this method

		.. latex:vspace:: -5px

		:param formula: A string with a chemical formula
		:param charge:

		.. TODO:: should throw error for unrecognised elements CGCGAATTCGCG
		"""

		_class = cls()

		formula = str(formula)
		formula = formula.strip().replace(' ', '')

		# Substitute abbreviations of common chemical groups
		for grp in reversed(sorted(GROUPS)):
			formula = formula.replace(grp, f"({GROUPS[grp]})")

		comp_and_charge = string_to_composition(formula)
		# print(comp_and_charge)

		if 0 in comp_and_charge:
			if charge:
				if comp_and_charge[0] != charge:
					raise ValueError("Cannot supply 'charge' when the formula already has a charge!")

			charge = comp_and_charge[0]

		for symbol, number in comp_and_charge.items():
			if number == 0:
				raise ValueError(f"Unrecognised formula: {formula}")
			# isotope = 0
			if symbol == 0:
				continue

			# isotope_re_1 = re.match(r"([A-z]+)(\[[0-9]*\])", symbol)
			# isotope_re_2 = re.findall(r"(\[[A-z]+)([0-9]*\])", symbol)
			#
			# if isotope_re_1:
			# 	elem, isotope = isotope_re_1[0]
			# 	isotope = re.findall(r"(\[)([0-9]+)(\])", isotope)[0][1]
			# 	if not isotope:
			# 		isotope = 0
			# elif isotope_re_2:
			# 	elem, isotope = isotope_re_2[0]
			# 	elem = elem.lstrip("[")
			# 	isotope = isotope.rstrip("]")
			# elif symbol not in ELEMENTS:
			# 	raise ValueError(f'Unknown chemical element with symbol {symbol}')
			# else:
			# 	elem = ELEMENTS[symbol].symbol
			elem, isotope = split_isotope(symbol)

			iso_str = _make_isotope_string(elem, int(isotope) if isotope else 0)
			_class[iso_str] += int(number) if number else 1

		_class._set_charge(charge)
		return _class

	@classmethod
	def from_mass_fractions(
			cls: Type['F'],
			fractions: Dict[str, float],
			charge: int = 0,
			maxcount: int = 10,
			precision: float = 1e-4,
			) -> "Formula":
		"""
		Create a new :class:`~chemistry_tools.formulae.formula.Formula` object
		from elemental mass fractions by parsing a string.

		.. latex:vspace:: -5px

		.. note:: Isotopes cannot (currently) be parsed using this method

		.. latex:vspace:: -5px

		:param fractions: A dictionary of elements and mass fractions
		:param charge:
		:param maxcount:
		:param precision:

		:bold-title:`Examples:`

		.. code-block:: python

			>>> Formula.from_mass_fractions({'H': 0.112, 'O': 0.888})
			'H2O'
			>>> Formula.from_mass_fractions({'D': 0.2, 'O': 0.8})
			'O[2H]2'
			>>> Formula.from_mass_fractions({'H': 8.97, 'C': 59.39, 'O': 31.64})
			'C5H9O2'
			>>> Formula.from_mass_fractions({'O': 0.26, '30Si': 0.74})
			'O2[30Si]3'
		"""  # noqa: D400

		# divide normalized fractions by element/isotope mass
		numbers = {}
		sumfractions = sum(fractions.values())
		for symbol, fraction in fractions.items():
			if symbol[0].isupper():
				try:
					mass = ELEMENTS[symbol].mass
				except KeyError as exc:
					raise ValueError(f"Unknown element '{symbol}'") from exc
			else:
				if symbol.startswith('['):
					symbol = symbol[1:-1]
				i = 0
				while symbol[i].isdigit():
					i += 1
				massnum = int(symbol[:i])
				symbol = symbol[i:]
				try:
					mass = ELEMENTS[symbol].isotopes[massnum].mass
				except KeyError as exc:
					raise ValueError(f"Unknown isotope '[{massnum}{symbol}]'") from exc
				symbol = f"[{massnum}{symbol}]"
			numbers[symbol] = fraction / (sumfractions * mass)

		# divide numbers by smallest number
		smallest = min(numbers.values())
		for symbol in numbers:
			numbers[symbol] /= smallest

		# find smallest factor that turns all numbers into integers
		precision *= len(numbers)
		best = 1e6
		factor = 1
		for i in range(1, maxcount):
			x = sum(abs((i * n) - round(i * n)) for n in numbers.values())
			if x < best:
				best = x
				factor = i
				if best < i * precision:
					break

		formula = []
		for symbol, number in sorted(numbers.items()):
			count = int(round(factor * number))
			if count > 1:
				formula.append(f'{symbol}{count}')
			else:
				formula.append(symbol)

		return cls.from_string(''.join(formula), charge=charge)

	@classmethod
	def from_kwargs(cls: Type['F'], *, charge: int = 0, **kwargs) -> F:
		"""
		Create a new :class:`~chemistry_tools.formulae.formula.Formula` object from
		keyword arguments representing the elements in the compound.

		:param charge:

		:rtype:

		.. latex:clearpage::
		"""  # noqa: D400

		return cls(kwargs, charge=charge)

	@property
	def monoisotopic_mass(self) -> float:
		"""
		Calculate the monoisotopic mass of a :class:`~chemistry_tools.formulae.formula.Formula`.
		If any isotopes are already present in the formula, the mass of these will be preserved
		"""

		# Calculate mass
		mass = 0.0

		for element, count in self.items():
			if element == 'D':
				iso_mass = D.mass
			elif element == 'T':
				iso_mass = T.mass
			else:
				symbol, isotope = split_isotope(element)

				if isotope:
					iso_mass = ELEMENTS[symbol].isotopes[isotope].mass
				else:
					iso_mass = ELEMENTS[symbol].isotopes[ELEMENTS[symbol].nominalmass].mass

			mass += iso_mass * count

		return mass

	@property
	def nominal_mass(self) -> float:
		"""
		Calculate the monoisotopic mass of a :class:`~chemistry_tools.formulae.formula.Formula`.
		If any isotopes are already present in the formula, the mass of these will be preserved
		"""

		return self.monoisotopic_mass

	@property
	def exact_mass(self) -> float:
		"""
		Calculate the monoisotopic mass of a :class:`~chemistry_tools.formulae.formula.Formula`.
		If any isotopes are already present in the formula, the mass of these will be preserved
		"""

		return self.monoisotopic_mass

	@property
	def mass(self) -> float:
		"""
		Calculate the average mass of a :class:`~chemistry_tools.formulae.formula.Formula`.

		Note that mass is not averaged for elements with specified isotopes.
		"""

		# Calculate mass
		mass = 0.0

		for element, count in self.items():
			if element == 'D':
				iso_mass = D.mass
			elif element == 'T':
				iso_mass = T.mass
			else:
				symbol, isotope = split_isotope(element)

				if isotope:
					iso_mass = ELEMENTS[symbol].isotopes[isotope].mass
				else:
					iso_mass = ELEMENTS[symbol].mass

			mass += iso_mass * count

		return mass

	@property
	def average_mass(self) -> float:
		"""
		Calculate the average mass of a :class:`~chemistry_tools.formulae.formula.Formula`.

		Note that mass is not averaged for elements with specified isotopes.
		"""

		return self.mass

	@property
	def mz(self) -> float:
		"""
		The mass to charge ratio of the formula.
		"""

		return self.get_mz(average=False)

	@property
	def average_mz(self) -> float:
		"""
		The average mass to charge ratio of the formula.
		"""

		return self.get_mz(average=True)

	def get_mz(self, average: bool = True, charge: Optional[int] = None) -> float:
		"""
		Calculate the average mass:charge ratio (*m/z*) of a :class:`~chemistry_tools.formulae.formula.Formula`.

		:param average: If :py:obj:`True` then the average *m/z* is calculated. Note that the mass
			is not averaged for elements with specified isotopes.
		:param charge: The charge of the compound. If :py:obj:`None` then the existing charge of the Formula is used
		"""

		if average:
			mass = self.mass
		else:
			mass = self.monoisotopic_mass

		# Calculate m/z
		if charge:
			mass /= charge
		elif self.charge:
			mass /= self.charge

		return mass

	def most_probable_isotopic_composition(
			self,
			elements_with_isotopes: Optional[Sequence[str]] = None,
			) -> Tuple["Formula", float]:
		"""
		Calculate the most probable isotopic composition of a molecule/ion.

		For each element, only two most abundant isotopes are considered.
		Any isotopes already in the Formula will be changed to the most abundant isotope

		:param elements_with_isotopes: A set of elements to be considered in isotopic distribution
			(by default, every element has an isotopic distribution).

		:return: A tuple with the most probable isotopic composition and its
			relative abundance.
		"""

		# Removing isotopes from the composition.
		for isotope_string in self:
			element_name, isotope_num = split_isotope(isotope_string)
			if isotope_num:
				self[element_name] += self.pop(isotope_string)

		isotopic_composition = Formula()

		for element_name in self:
			if not elements_with_isotopes or (element_name in elements_with_isotopes):
				# Take the two most abundant isotopes.
				first_iso, second_iso = sorted(
					[(i[0], i[1][1]) for i in isotope_data[element_name].items() if i[0]],
					key=lambda x: -x[1]
					)[:2]

				# Write the number of isotopes of the most abundant type.
				first_iso_str = _make_isotope_string(element_name, first_iso[0])
				isotopic_composition[first_iso_str] = int(math.ceil(self[element_name])) * first_iso[1]

				# Write the number of the second isotopes.
				second_iso_str = _make_isotope_string(element_name, second_iso[0])
				isotopic_composition[second_iso_str] = self[element_name] - isotopic_composition[first_iso_str]
			else:
				isotopic_composition[element_name] = self[element_name]

		return isotopic_composition, isotopic_composition.isotopic_composition_abundance

	@property
	def isotopic_composition_abundance(self) -> float:
		"""
		Calculate the relative abundance of the current isotopic composition of this molecule.

		:returns: The relative abundance of the current isotopic composition.
		"""

		isotopic_composition: defaultdict = defaultdict(dict)

		# Check if there are default and non-default isotopes of the same
		# element and rearrange the elements.
		for element in self:
			element_name, isotope_num = split_isotope(element)

			# If there is already an entry for this element and either it
			# contains a default isotope or newly added isotope is default
			# then raise an exception.
			if (
					element_name in isotopic_composition
					and (isotope_num == 0 or 0 in isotopic_composition[element_name])
					):
				raise ValueError(
						"Please specify the isotopic states of all atoms of "
						f"{element_name} or do not specify them at all."
						)
			else:
				isotopic_composition[element_name][isotope_num] = (self[element])

		# Calculate relative abundance.
		num1, num2, denom = 1, 1, 1

		for element_name, isotope_dict in isotopic_composition.items():
			num1 *= math.factorial(sum(isotope_dict.values()))
			for isotope_num, isotope_content in isotope_dict.items():
				denom *= math.factorial(isotope_content)
				if isotope_num:
					num2 *= (isotope_data[element_name][isotope_num][1]**isotope_content)

		return num2 * (num1 / denom)

	def iter_isotopologues(
			self,
			report_abundance: bool = False,
			elements_with_isotopes: Optional[Sequence[str]] = None,
			isotope_threshold: float = 5e-4,
			overall_threshold: float = 0,
			) -> Iterator:  # TODO: of what?
		"""
		Iterate over possible isotopic states of the molecule.

		The space of possible isotopic compositions is restrained by parameters
		``elements_with_isotopes``, ``isotope_threshold``, ``overall_threshold``.

		:param report_abundance: If :py:obj:`True`, the output will contain 2-tuples: `(composition, abundance)`.
			Otherwise, only compositions are yielded.
		:param elements_with_isotopes: A set of elements to be considered in isotopic distributions
			(by default, every element has an isotopic distribution).
		:param isotope_threshold: The threshold abundance of a specific isotope to be considered.
		:param overall_threshold: The threshold abundance of the calculated isotopic composition.

		:return: Iterator over possible isotopic compositions.

		.. latex:clearpage::
		"""

		dict_elem_isotopes = {}
		for element in self:
			if elements_with_isotopes is None or element in elements_with_isotopes:
				element_name, isotope_num = split_isotope(element)
				isotopes = {
					k: v
					for k, v in isotope_data[element_name].items()
					if k != 0 and v[1] >= isotope_threshold}  # yapf: disable
				list_isotopes = [_make_isotope_string(element_name, k) for k in isotopes]
				dict_elem_isotopes[element] = list_isotopes
			else:
				dict_elem_isotopes[element] = [element]

		all_isotoplogues = []
		for element, list_isotopes in dict_elem_isotopes.items():
			n = self[element]
			list_comb_element_n = []
			for elementXn in combinations_with_replacement(list_isotopes, n):
				list_comb_element_n.append(elementXn)
			all_isotoplogues.append(list_comb_element_n)

		for isotopologue in product(*all_isotoplogues):
			flat_isotopologue = [atom for element in isotopologue for atom in element]
			ic = Formula(Counter(flat_isotopologue))
			if report_abundance or overall_threshold > 0.0:
				abundance = ic.isotopic_composition_abundance
				if abundance > overall_threshold:
					if report_abundance:
						yield ic, abundance
					else:
						yield ic
			else:
				yield ic

	def isotope_distribution(self) -> IsotopeDistribution:
		"""
		Returns an :class:`~.IsotopeDistribution` object representing the distribution of the
		isotopologues of the formula.
		"""  # noqa: D400

		return IsotopeDistribution(self)

	def copy(self: F) -> F:
		"""
		Returns a copy of the :class:`~.Formula`.
		"""

		return self.__class__(self, charge=self.charge)

	def __missing__(self, key) -> int:  # noqa: MAN001
		# override default behavior: we don't want to add 0's to the dictionary
		return 0

	def __setitem__(self, key: str, value: float) -> None:
		if isinstance(value, float):
			value = int(round(value))
		elif not isinstance(value, int):
			raise TypeError(f"Only integers allowed as values in Formula, got {type(value).__name__}.")
		if value:  # reject 0's
			super(defaultdict, self).__setitem__(key, value)
		elif key in self:
			del self[key]

	def __add__(self, other):  # noqa: MAN001,MAN002
		result = self.copy()
		for elem, count in other.items():
			result[elem] += count
		return result

	def __iadd__(self, other) -> "Formula":  # noqa: MAN001
		for elem, count in other.items():
			self[elem] += count
		return self

	def __radd__(self, other):  # noqa: MAN001,MAN002
		return self + other

	def __sub__(self, other):  # noqa: MAN001,MAN002
		result = self.copy()
		for elem, count in other.items():
			result[elem] -= count
		return result

	def __isub__(self, other) -> "Formula":  # noqa: MAN001
		for elem, count in other.items():
			self[elem] -= count
		return self

	def __rsub__(self, other):  # noqa: MAN001,MAN002
		return (self - other) * (-1)

	def __mul__(self, other) -> "Formula":  # noqa: MAN001
		if not isinstance(other, int):
			raise TypeError(f'Cannot multiply Formula by non-integer "{other}"')
		return type(self)({k: v * other for k, v in self.items()})

	def __imul__(self, other) -> "Formula":  # noqa: MAN001
		if not isinstance(other, int):
			raise TypeError(f'Cannot multiply Formula by non-integer "{other}"')
		for elem in self:
			self[elem] *= other
		return self

	def __rmul__(self, other):  # noqa: MAN001,MAN002
		return self * other

	def __eq__(self, other) -> bool:  # noqa: MAN001
		if isinstance(other, (dict, Formula)):
			self_items = {i for i in self.items() if i[1]}
			other_items = {i for i in other.items() if i[1]}

			if isinstance(other, Formula):
				return self_items == other_items and self.charge == other.charge
			else:
				return self_items == other_items
		else:
			return NotImplemented

	def __str__(self) -> str:
		return self.hill_formula

	def _repr_elements(self) -> List[str]:
		elements = [str(dict.__repr__(self))]

		if self.charge:
			elements.append(f"charge={self.charge}")

		return elements

	def __repr__(self) -> str:
		return f'{type(self).__name__}({", ".join(self._repr_elements())})'

	@property
	def hill_formula(self) -> str:
		"""
		Returns the formula in Hill notation.

		:bold-title:`Example:`

		.. code-block:: python

			>>> Formula.from_string('BrC2H5').hill_formula
			'C2H5Br'
			>>> Formula.from_string('HBr').hill_formula
			'BrH'
			>>> Formula.from_string('[(CH3)3Si2]2NNa').hill_formula
			'C6H18NNaSi4'
		"""

		hill = []

		for symbol in hill_order(self.elements):
			hill.append(symbol)
			count = self[symbol]
			if count > 1:
				hill.append(str(count))

		# alphabet = sorted(ELEMENTS.symbols)
		#
		# if "C" in self:
		# 	hill.append("C")
		# 	alphabet.remove("C")
		# 	count = self["C"]
		# 	if count > 1:
		# 		hill.append(str(count))
		# 	if "H" in self:
		# 		hill.append("H")
		# 		alphabet.remove("H")
		# 		count = self["H"]
		# 		if count > 1:
		# 			hill.append(str(count))
		#
		# for element in alphabet:
		# 	if element in self:
		# 		hill.append(element)
		# 		count = self[element]
		# 		if count > 1:
		# 			hill.append(str(count))
		#
		return ''.join(hill)

	@property
	def no_isotope_hill_formula(self) -> str:
		"""
		Returns formula in Hill notation, without any isotopes specified.

		:bold-title:`Example:`

		.. code-block:: python

			>>> Formula.from_string('BrC2H5').no_isotope_hill_formula
			'C2H5Br'
			>>> Formula.from_string('HBr').no_isotope_hill_formula
			'BrH'
			>>> Formula.from_string('[(CH3)3Si2]2NNa').no_isotope_hill_formula
			'C6H18NNaSi4'
		"""

		hill = []

		ordered_symbols: Dict[str, int] = dict()

		for symbol in hill_order(self.elements):
			count = self[symbol]
			symbol, _ = split_isotope(symbol)
			if symbol in ordered_symbols:
				ordered_symbols[symbol] += count
			else:
				ordered_symbols[symbol] = count

		for symbol, count in ordered_symbols.items():
			hill.append(symbol)
			if count > 1:
				hill.append(str(count))

		return ''.join(hill)

	@property
	def empirical_formula(self) -> str:
		"""
		Returns the empirical formula in Hill notation.

		The empirical formula has the simplest whole number ratio of atoms
		of each element present in the formula.

		:bold-title:`Examples:`

		.. code-block:: python

			>>> Formula.from_string('H2O').empirical_formula
			'H2O'
			>>> Formula.from_string('S4').empirical_formula
			'S'
			>>> Formula.from_string('C6H12O6').empirical_formula
			'CH2O'
		"""

		hill = []
		values = list(self.values())
		if len(values) > 1:
			divisor = gcd_array(values)
		else:
			divisor = values[0]

		for symbol in hill_order(self.elements):
			hill.append(symbol)
			count = self[symbol] // divisor
			if count > 1:
				hill.append(str(count))

		return ''.join(hill)

	@property
	def n_atoms(self) -> int:
		"""
		Return the number of atoms in the formula.

		:bold-title:`Example:`

		.. code-block:: python

			>>> Formula.from_string('CH3COOH').n_atoms
			8
		"""

		return sum(list(self.values()))

	@property
	def n_elements(self) -> int:
		"""
		Return the number of elements in the formula.

		:rtype:

		:bold-title:`Example:`

		.. code-block:: python

			>>> Formula.from_string('CH3COOH').n_elements
			3


		.. latex:clearpage::
		"""

		return len(self)

	@property
	def elements(self) -> List[str]:  # type: ignore[override]
		"""
		A list of the element symbols in the formula.
		"""

		return list(self.keys())

	@property
	def composition(self) -> "Composition":
		"""
		A :class:`~.Composition` object representing the elemental composition of the Formula.
		"""

		return Composition(self)



#!/usr/bin/env python3
#
#  dataarray.py
"""
Provides a base class which can output data as a :class:`pandas.DataFrame`,
to CSV, or as a pretty-printed table in a variety of formats.
"""  # noqa: D400
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

# stdlib
from abc import abstractmethod
#  3rd party
from typing import Any, Dict, List, Optional, Set

# 3rd party
import pandas  # type: ignore[import]
import tabulate  # nodep
from cawdrey import FrozenOrderedDict  # nodep
from domdf_python_tools.doctools import prettify_docstrings

__all__ = ["DataArray"]


@prettify_docstrings
class DataArray(FrozenOrderedDict):
	"""
	A class which can output data as a :class:`pandas.DataFrame`, to CSV,
	or as a pretty-printed table in a variety of formats.

	To use this class it must first be subclassed.
	Subclasses must implement :meth:`~DataArray.as_array` which handles the
	conversion of the data to a list of lists of values.

	:param formula: The formula in hill notation
	:param data: A dictionary of data to add to the internal :class:`~cawdrey.FrozenOrderedDict`
	"""  # noqa: D400

	_as_array_kwargs: Set[str] = set()
	_as_table_alignment: List[str] = []
	_as_table_float_format: List[Optional[str]] = []

	def __init__(self, formula: str, data: Dict):
		super().__init__(**data)
		self.formula: str = formula

	def as_csv(self, *args, sep: str = ',', **kwargs) -> str:
		r"""
		Returns the data as a CSV formatted string.

		:param \*args: Arguments passed to :meth:`~.DataArray.as_array`.
		:param sep: The separator for the CSV data.
		:param \*\*kwargs: Additional keyword arguments passed to :meth:`~.DataArray.as_array`.
		"""

		return '\n'.join(sep.join(x) for x in self.as_array(*args, **kwargs))

	@abstractmethod
	def as_array(self, sort_by: Any, reverse: bool = False) -> List[List[Any]]:
		"""
		Must be implemented in subclasses to hand the conversion of the data to a list of lists of values.

		:param sort_by:
		:param reverse:
		"""

	def as_dataframe(self, *args, **kwargs) -> pandas.DataFrame:
		"""
		Returns the isotope distribution data as a :class:`pandas.DataFrame`.

		Any arguments taken by :meth:`~.DataArray.as_array` can also be used here.
		"""

		array = self.as_array(*args, **kwargs)
		return pandas.DataFrame.from_records(array[1:], columns=array[0])

	def as_table(self, *args, **kwargs) -> str:
		"""
		Returns the isotope distribution data as a table using
		`tabulate <https://github.com/astanin/python-tabulate>`_.

		Any arguments taken by :meth:`~.DataArray.as_array` can also be used here.

		Additionally, any valid keyword argument for :func:`tabulate.tabulate` can be used.
		"""  # noqa: D400

		tabulate_kwargs: Dict[str, Any] = {}
		array_kwargs: Dict[str, Any] = {}

		for arg, val in kwargs.items():
			if arg in self._as_array_kwargs:
				array_kwargs[arg] = val
			else:
				tabulate_kwargs[arg] = val

		if "colalign" not in tabulate_kwargs:
			tabulate_kwargs["colalign"] = self._as_table_alignment

		if "floatfmt" not in tabulate_kwargs:
			tabulate_kwargs["floatfmt"] = self._as_table_float_format

		array = self.as_array(*args, **array_kwargs)

		return tabulate.tabulate(array[1:], array[0], **tabulate_kwargs)

	def __str__(self) -> str:
		return self.__repr__()

	def __repr__(self) -> str:
		return f"<{self.__class__.__name__}({self.formula})>"



#!/usr/bin/env python3
#
#  isotope_masses.py
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on Pyteomics (https://github.com/levitsky/pyteomics)
#  |  Copyright (c) 2011-2015, Anton Goloborodko & Lev Levitsky
#  |  Licensed under the Apache License, Version 2.0 (the "License");
#  |  you may not use this file except in compliance with the License.
#  |  You may obtain a copy of the License at
#  |
#  |    http://www.apache.org/licenses/LICENSE-2.0
#  |
#  |  Unless required by applicable law or agreed to in writing, software
#  |  distributed under the License is distributed on an "AS IS" BASIS,
#  |  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  |  See the License for the specific language governing permissions and
#  |  limitations under the License.
#  |
#  |  See also:
#  |  Goloborodko, A.A.; Levitsky, L.I.; Ivanov, M.V.; and Gorshkov, M.V. (2013)
#  |  "Pyteomics - a Python Framework for Exploratory Data Analysis and Rapid Software
#  |  Prototyping in Proteomics", Journal of The American Society for Mass Spectrometry,
#  |  24(2), 301–304. DOI: `10.1007/s13361-012-0516-6 <http://dx.doi.org/10.1007/s13361-012-0516-6>`_
#  |
#  |  Levitsky, L.I.; Klein, J.; Ivanov, M.V.; and Gorshkov, M.V. (2018)
#  |  "Pyteomics 4.0: five years of development of a Python proteomics framework",
#  |  Journal of Proteome Research.
#  |  DOI: `10.1021/acs.jproteome.8b00717 <http://dx.doi.org/10.1021/acs.jproteome.8b00717>`_
#

# stdlib
from typing import Dict, Tuple

# this package
from chemistry_tools.elements import ELEMENTS

isotopes_dict = Dict[int, Tuple[float, float]]

h_plus: isotopes_dict = {0: (1.00727646677, 1.0), 1: (1.00727646677, 1.0)}
electron: isotopes_dict = {0: (0.00054857990943, 1.0)}

isotope_data = dict()

for element in ELEMENTS:

	# if element.symbol in undefined_isotopes:
	# 	element_isotopes = undefined_isotopes[element.symbol]
	# else:
	element_isotopes = {}

	for massnumber, isotope in element.isotopes.items():
		element_isotopes[massnumber] = (isotope.mass, isotope.abundance)
		element_isotopes[0] = (element.mass, 1.0)

	isotope_data[element.symbol] = element_isotopes
	assert isotope_data[element.symbol][0]

isotope_data["H+"] = h_plus
isotope_data["e*"] = electron



#!/usr/bin/env python3
#
#  _parser_core.py
"""
Core functions and constants for parsing formulae.
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on Pyteomics (https://github.com/levitsky/pyteomics)
#  |  Copyright (c) 2011-2015, Anton Goloborodko & Lev Levitsky
#  |  Licensed under the Apache License, Version 2.0 (the "License");
#  |  you may not use this file except in compliance with the License.
#  |  You may obtain a copy of the License at
#  |
#  |    http://www.apache.org/licenses/LICENSE-2.0
#  |
#  |  Unless required by applicable law or agreed to in writing, software
#  |  distributed under the License is distributed on an "AS IS" BASIS,
#  |  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  |  See the License for the specific language governing permissions and
#  |  limitations under the License.
#  |
#  |  See also:
#  |  Goloborodko, A.A.; Levitsky, L.I.; Ivanov, M.V.; and Gorshkov, M.V. (2013)
#  |  "Pyteomics - a Python Framework for Exploratory Data Analysis and Rapid Software
#  |  Prototyping in Proteomics", Journal of The American Society for Mass Spectrometry,
#  |  24(2), 301–304. DOI: `10.1007/s13361-012-0516-6 <http://dx.doi.org/10.1007/s13361-012-0516-6>`_
#  |
#  |  Levitsky, L.I.; Klein, J.; Ivanov, M.V.; and Gorshkov, M.V. (2018)
#  |  "Pyteomics 4.0: five years of development of a Python proteomics framework",
#  |  Journal of Proteome Research.
#  |  DOI: `10.1021/acs.jproteome.8b00717 <http://dx.doi.org/10.1021/acs.jproteome.8b00717>`_
#
#  Also based on ChemPy (https://github.com/bjodah/chempy)
#  |  Copyright (c) 2015-2018, Björn Dahlgren
#  |  All rights reserved.
#  |
#  |  Redistribution and use in source and binary forms, with or without modification,
#  |  are permitted provided that the following conditions are met:
#  |
#  |    Redistributions of source code must retain the above copyright notice, this
#  |    list of conditions and the following disclaimer.
#  |
#  |    Redistributions in binary form must reproduce the above copyright notice, this
#  |    list of conditions and the following disclaimer in the documentation and/or
#  |    other materials provided with the distribution.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  |  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  |  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  |  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
#  |  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  |  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  |  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#  |  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  |  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  |  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

# stdlib
import re
import warnings
from typing import Any, Callable, Dict, Iterable, List, Optional, Sequence, Tuple, Union

__all__ = ["replace_substrings"]

_greek_letters: Tuple[str, ...] = (
		"alpha",
		"beta",
		"gamma",
		"delta",
		"epsilon",
		"zeta",
		"eta",
		"theta",
		"iota",
		"kappa",
		"lambda",
		"mu",
		"nu",
		"xi",
		"omicron",
		"pi",
		"rho",
		"sigma",
		"tau",
		"upsilon",
		"phi",
		"chi",
		"psi",
		"omega",
		)

_greek_u = "αβγδεζηθικλμνξοπρστυφχψω"


def _formula_to_format(
		sub: Callable,
		sup: Callable,
		formula: str,
		prefixes: Dict[str, str],
		infixes: Dict[str, str],
		suffixes: Sequence[str] = ("(s)", "(l)", "(g)", "(aq)"),
		) -> str:
	"""

	:param sub: The function to call to subscript a string.
	:param sup: The function to call to superscript a string.
	:param formula: The formula to format
	:param prefixes: Mapping of prefixes to their equivalents in the desired format
	:param infixes: Mapping of infixes to their equivalents in the desired format
	:param suffixes: Suffixes to keep, e.g. ("(g)", "(s)")

	:return: The formatted formula
	"""

	# TODO: make isotope square brackets be superscript
	parts = _formula_to_parts(formula, prefixes.keys(), suffixes)
	stoichs = parts[0].split('.')
	string = ''

	for idx, stoich in enumerate(stoichs):
		if idx == 0:
			m = 1
		else:
			m, stoich = _get_leading_integer(stoich)
			string += replace_substrings('.', infixes)
		if m != 1:
			string += str(m)

		string += re.sub(r"([0-9]+)(?![^\[]*\])", lambda m: sub(m.group(1)), stoich)

	if parts[1] is not None:
		chg = _get_charge(parts[1])
		if chg < 0:
			token = '-' if chg == -1 else f"{-chg:d}-"
		if chg > 0:
			token = '+' if chg == 1 else f"{chg:d}+"
		string += sup(token)

	if len(parts) > 4:
		raise ValueError("Incorrect formula")

	pre_str = ''.join(map(lambda x: replace_substrings(x, prefixes), parts[2]))

	return pre_str + string + ''.join(parts[3])


def _formula_to_parts(
		formula: str,
		prefixes: Iterable[str],
		suffixes: Sequence[str] = ("(s)", "(l)", "(g)", "(aq)"),
		) -> List[Any]:
	"""

	:param formula: The formula to split into parts.
	:param prefixes: Mapping of prefixes to their HTML equivalents. Default greek letters and ``.``
	:param suffixes: Suffixes to keep, e.g. ("(g)", "(s)")
	"""

	# Drop prefixes and suffixes
	drop_pref, drop_suff = [], []

	for ign in prefixes:
		if formula.startswith(ign):
			drop_pref.append(ign)
			formula = formula[len(ign):]

	for ign in suffixes:
		if formula.endswith(ign):
			drop_suff.append(ign)
			formula = formula[:-len(ign)]

	parts: Tuple[Optional[str], ...]

	# Extract charge
	for token in "+-":
		if token in formula:
			if formula.count(token) > 1:
				raise ValueError(f"Multiple tokens: {token}")
			parts_: List[str] = formula.split(token)
			parts = (parts_[0], token + parts_[1], *parts_[2:])
			break
	else:
		parts = (formula, None)

	return [*parts, tuple(drop_pref), tuple(drop_suff[::-1])]


def replace_substrings(string: str, patterns: Dict[str, str]) -> str:
	"""
	Replace substrings in a string.

	:param string: The string to replace substrings in.
	:param patterns: A dictionary mapping substrings to their replacements.

	:return: The resulting string.
	"""

	for patt, repl in patterns.items():
		string = string.replace(patt, repl)

	return string


def _get_leading_integer(s: str) -> Tuple[int, str]:
	"""
	Returns the leading integer from the string. If no leading integer is found it is assumed to be ``1``.

	:param s: The string to parse

	:return: A tuple comprising the leading integer and the remainder of the string
	"""

	matches = re.findall(r"^\d+", s)

	if len(matches) == 0:
		integer = 1
	elif len(matches) == 1:
		s = s[len(matches[0]):]
		integer = int(matches[0])
	else:
		raise ValueError(f"Failed to parse: {s}")

	return integer, s


def _get_charge(charge_str: str) -> int:
	"""
	Parses a string representing a charge.

	:param charge_str:

	:return: The charge

	:raises ValueError: If the charge string cannot be parsed.
	"""

	if charge_str == '+':
		return 1
	elif charge_str == '-':
		return -1

	for token, anti, sign in zip("+-", "-+", (1, -1)):
		if token in charge_str:
			if anti in charge_str:
				raise ValueError("Invalid charge description (+ & - present)")

			before, after = charge_str.split(token)

			if len(before) > 0 and len(after) > 0:
				raise ValueError("Values both before and after charge token")

			if len(before) > 0:
				# will_be_missing_in="0.8.0'
				warnings.warn("'Fe/3+' deprecated, use e.g. 'Fe+3'", DeprecationWarning, stacklevel=3)
				return sign * int(1 if before == '' else before)

			if len(after) > 0:
				return sign * int(1 if after == '' else after)

	raise ValueError("Invalid charge description (+ or - missing)")


def _make_isotope_string(element_name: str, isotope_num: Union[str, int]) -> str:
	"""
	Form a string label for an isotope. If ``isotope_num`` = 0 ``element_name`` is returned unchanged.

	:param element_name: The name or symbol of the element
	:param isotope_num: The isotope number

	:return: The isotope string
	"""

	if isotope_num in {0, '0'}:
		return element_name
	else:
		return f"[{isotope_num}{element_name}]"


_isotope_string = re.compile(r"^([A-Z][a-z+]*)(?:\[(\d+)])?$")


# TODO: merge with split_isotope
def _parse_isotope_string(label: str) -> Tuple[str, int]:
	"""
	Parse an string with an isotope label and return the element name and the isotope number.

	>>> _parse_isotope_string("C")
	("C", 0)
	>>> _parse_isotope_string("C[12]")
	("C", 12)

	:param label: The isotope label to parse

	:return: The name/symbol of the element, and the isotope number
	"""

	matches = _isotope_string.match(label)
	if matches:
		element_name, num = matches.groups()
		isotope_num = int(num) if num else 0
		return element_name, isotope_num
	else:
		raise ValueError(f"Failed to parse: {label}")



#!/usr/bin/env python3
#
#  iso_dist.py
"""
Isotope Distributions.
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

# stdlib
from collections import OrderedDict
from typing import Any, List, Union

# 3rd party
from domdf_python_tools.doctools import prettify_docstrings
from enum_tools import IntEnum
from enum_tools.documentation import document_enum

# this package
from chemistry_tools import formulae

# this package
from .dataarray import DataArray
from .unicode import string_to_unicode

__all__ = ["IsoDistSort", "IsotopeDistribution"]


@document_enum
class IsoDistSort(IntEnum):
	"""
	Lookup for sorting isotope distribution output.
	"""

	Formula = formula = 0  # doc: Sort the isosope distribution by the formulae.
	Mass = mass = 1  # doc: Sort the isotope distribution by the masses.
	Abundance = abundance = 2  # doc: Sort the isotope distribution by the abundances.
	Relative_Abundance = Relative_abundance = relative_abundance = 3  # doc: Sort the isotope distribution by the relative abundances.


@prettify_docstrings
class IsotopeDistribution(DataArray):
	"""
	An isotope distribution.

	:param formula: A :class:`~chemistry_tools.formulae.formula.Formula` object to create the distribution for

	Each composition can be accessed with their hill formulae like a dictionary
	(e.g. ``iso_dict['H[1]2O[16]']``)
	"""

	# TODO: as_mass_spec

	def __init__(self, formula: "formulae.Formula"):
		iso_compositions = list(formula.iter_isotopologues())
		compositions = OrderedDict()
		max_abundance: float = 0

		for comp in iso_compositions:
			compositions[comp.hill_formula] = comp
			abundance = comp.isotopic_composition_abundance
			if abundance > max_abundance:
				max_abundance = abundance

		super().__init__(formula=iso_compositions[0].no_isotope_hill_formula, data=compositions)

		self.max_abundance: float = max_abundance

	_as_array_kwargs = {"sort_by", "reverse", "format_percentage"}
	_as_table_alignment = ["left", "right", "right", "right"]
	_as_table_float_format = [None, ".4f", ".6f", ".6f"]

	def as_array(
			self,
			sort_by: Union[int, IsoDistSort] = IsoDistSort.formula,
			reverse: bool = False,
			format_percentage: bool = True
			) -> List[List[Any]]:
		"""
		Returns the isotope distribution data as a list of lists.

		:param sort_by: The column to sort by.
		:param reverse: Whether the isotopologues should be sorted in reverse order.
		:param format_percentage: Whether the abundances should be formatted as percentages or not.
		"""

		if sort_by == IsoDistSort.formula:
			sort_key = lambda comp: comp.hill_formula
		elif sort_by == IsoDistSort.mass:
			sort_key = lambda comp: comp.mass
		elif sort_by == IsoDistSort.abundance:
			sort_key = lambda comp: comp.isotopic_composition_abundance
		elif sort_by == IsoDistSort.relative_abundance:
			sort_key = lambda comp: comp.isotopic_composition_abundance / self.max_abundance
		else:
			raise ValueError(f"Unrecognised value for 'sort_by': {sort_by}")

		output = []

		for comp in sorted(self.values(), key=sort_key, reverse=reverse):
			abundance = comp.isotopic_composition_abundance
			rel_abund = abundance / self.max_abundance
			row = [comp.hill_formula, f"{comp.mass:0.4f}"]
			if format_percentage:
				row += [f"{abundance:0.2%}", f"{rel_abund:0.2%}"]
			else:
				row += [f"{abundance:0.6f}", f"{rel_abund:0.6f}"]
			output.append(row)
			# TODO: Unicode, latex, html representations of formulae

		output.insert(0, ["Formula", "Mass", "Abundance", "Relative Abundance"])

		return output

	def __str__(self) -> str:
		table = self.as_table(sort_by=IsoDistSort.relative_abundance, reverse=True, tablefmt="fancy_grid")
		return f"\n Isotope Distribution for {string_to_unicode(self.formula)}\n{table}"



#!/usr/bin/env python3
#
#  pnictogens.py
r"""
Group 15: Pnictogens in the Periodic Table.

.. raw:: latex

	\begin{multicols}{2}

.. data:: N

	:class:`~chemistry_tools.elements.classes.Element` representing Nitrogen

.. data:: P

	:class:`~chemistry_tools.elements.classes.Element` representing Phosphorus

.. data:: As

	:class:`~chemistry_tools.elements.classes.Element` representing Arsenic

.. data:: Sb

	:class:`~chemistry_tools.elements.classes.Element` representing Antimony

.. data:: Bi

	:class:`~chemistry_tools.elements.classes.Element` representing Bismuth

.. data:: Mc

	:class:`~chemistry_tools.elements.classes.Element` representing Moscovium

.. raw:: latex

	\end{multicols}
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on molmass (https://github.com/cgohlke/molmass)
#  |  Copyright (c) 1990-2020, Christoph Gohlke
#  |  All rights reserved.
#  |  Licensed under the BSD 3-Clause License
#  |  Redistribution and use in source and binary forms, with or without
#  |  modification, are permitted provided that the following conditions are met:
#  |
#  |  1. Redistributions of source code must retain the above copyright notice,
#  |     this list of conditions and the following disclaimer.
#  |
#  |  2. Redistributions in binary form must reproduce the above copyright notice,
#  |     this list of conditions and the following disclaimer in the documentation
#  |     and/or other materials provided with the distribution.
#  |
#  |  3. Neither the name of the copyright holder nor the names of its
#  |     contributors may be used to endorse or promote products derived from
#  |     this software without specific prior written permission.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  |  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  |  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  |  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  |  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  |  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  |  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  |  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  |  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  |  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  |  POSSIBILITY OF SUCH DAMAGE.
#  |
#  Also based on Pyteomics (https://github.com/levitsky/pyteomics)
#  |  Copyright (c) 2011-2015, Anton Goloborodko & Lev Levitsky
#  |  Licensed under the Apache License, Version 2.0 (the "License");
#  |  you may not use this file except in compliance with the License.
#  |  You may obtain a copy of the License at
#  |
#  |    http://www.apache.org/licenses/LICENSE-2.0
#  |
#  |  Unless required by applicable law or agreed to in writing, software
#  |  distributed under the License is distributed on an "AS IS" BASIS,
#  |  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  |  See the License for the specific language governing permissions and
#  |  limitations under the License.
#  |
#  |  See also:
#  |  Goloborodko, A.A.; Levitsky, L.I.; Ivanov, M.V.; and Gorshkov, M.V. (2013)
#  |  "Pyteomics - a Python Framework for Exploratory Data Analysis and Rapid Software
#  |  Prototyping in Proteomics", Journal of The American Society for Mass Spectrometry,
#  |  24(2), 301–304. DOI: `10.1007/s13361-012-0516-6 <http://dx.doi.org/10.1007/s13361-012-0516-6>`_
#  |
#  |  Levitsky, L.I.; Klein, J.; Ivanov, M.V.; and Gorshkov, M.V. (2018)
#  |  "Pyteomics 4.0: five years of development of a Python proteomics framework",
#  |  Journal of Proteome Research.
#  |  DOI: `10.1021/acs.jproteome.8b00717 <http://dx.doi.org/10.1021/acs.jproteome.8b00717>`_
#
#  Isotope data from http://www.nist.gov/pml/data/comp.cfm .

# this package
from .classes import Element

__all__ = ('N', 'P', "As", "Sb", "Bi", "Mc")

# Isotope 0 Key:
# mass of the most abundant isotope and 1.0 abundance.

N = Element(
		7,
		'N',
		"Nitrogen",
		group=15,
		period=2,
		block='p',
		series=1,
		mass=14.006703,
		eleneg=3.04,
		eleaffin=-0.07,
		covrad=0.75,
		atmrad=0.75,
		vdwrad=1.55,
		tboil=77.344,
		tmelt=63.15,
		density=1.17,
		eleconfig="[He] 2s2 2p3",
		oxistates="5, 4, 3, 2, -3*",
		ionenergy=(
				14.5341,
				39.601,
				47.488,
				77.472,
				97.888,
				522.057,
				667.029,
				),
		isotopes={
				10: (10.04165, 0.0),
				11: (11.02609, 0.0),
				12: (12.0186132, 0.0),
				13: (13.00573861, 0.0),
				14: (14.0030740048, 0.99636),
				15: (15.0001088982, 0.00364),
				16: (16.0061017, 0.0),
				17: (17.00845, 0.0),
				18: (18.014079, 0.0),
				19: (19.017029, 0.0),
				20: (20.02337, 0.0),
				21: (21.02711, 0.0),
				22: (22.03439, 0.0),
				23: (23.04122, 0.0),
				24: (24.05104, 0.0),
				25: (25.06066, 0.0),
				},
		description=(
				"Colourless, gaseous element which belongs to group 15 of the "
				"periodic table. Constitutes ~78% of the atmosphere and is an "
				"essential part of the ecosystem. Nitrogen for industrial purposes "
				"is acquired by the fractional distillation of liquid air. "
				"Chemically inactive, reactive generally only at high temperatures "
				"or in electrical discharges. It was discovered in 1772 by D. "
				"Rutherford."
				)
		)

P = Element(
		15,
		'P',
		"Phosphorus",
		group=15,
		period=3,
		block='p',
		series=1,
		mass=30.973761998,
		eleneg=2.19,
		eleaffin=0.7465,
		covrad=1.06,
		atmrad=1.23,
		vdwrad=1.8,
		tboil=553.0,
		tmelt=317.3,
		density=1.82,
		eleconfig="[Ne] 3s2 3p3",
		oxistates="5*, 3, -3",
		ionenergy=(
				10.4867,
				19.725,
				30.18,
				51.37,
				65.023,
				220.43,
				263.22,
				309.41,
				371.73,
				424.5,
				479.57,
				560.41,
				611.85,
				2816.943,
				3069.762,
				),
		isotopes={
				24: (24.03435, 0.0),
				25: (25.02026, 0.0),
				26: (26.01178, 0.0),
				27: (26.99923, 0.0),
				28: (27.992315, 0.0),
				29: (28.9818006, 0.0),
				30: (29.9783138, 0.0),
				31: (30.97376163, 1.0),
				32: (31.97390727, 0.0),
				33: (32.9717255, 0.0),
				34: (33.973636, 0.0),
				35: (34.9733141, 0.0),
				36: (35.97826, 0.0),
				37: (36.97961, 0.0),
				38: (37.98416, 0.0),
				39: (38.98618, 0.0),
				40: (39.9913, 0.0),
				41: (40.99434, 0.0),
				42: (42.00101, 0.0),
				43: (43.00619, 0.0),
				44: (44.01299, 0.0),
				45: (45.01922, 0.0),
				46: (46.02738, 0.0),
				},
		description=(
				"Non-metallic element belonging to group 15 of the periodic table. "
				"Has a multiple allotropic forms. Essential element for living "
				"organisms. It was discovered by Brandt in 1669."
				)
		)

As = Element(
		33,
		"As",
		"Arsenic",
		group=15,
		period=4,
		block='p',
		series=5,
		mass=74.921595,
		eleneg=2.18,
		eleaffin=0.814,
		covrad=1.2,
		atmrad=1.33,
		vdwrad=1.85,
		tboil=876.0,
		tmelt=1090.0,
		density=5.72,
		eleconfig="[Ar] 3d10 4s2 4p3",
		oxistates="5, 3*, -3",
		ionenergy=(
				9.7886,
				18.633,
				28.351,
				50.13,
				62.63,
				127.6,
				),
		isotopes={
				60: (59.99313, 0.0),
				61: (60.98062, 0.0),
				62: (61.9732, 0.0),
				63: (62.96369, 0.0),
				64: (63.95757, 0.0),
				65: (64.94956, 0.0),
				66: (65.94471, 0.0),
				67: (66.93919, 0.0),
				68: (67.93677, 0.0),
				69: (68.93227, 0.0),
				70: (69.93092, 0.0),
				71: (70.927112, 0.0),
				72: (71.926752, 0.0),
				73: (72.923825, 0.0),
				74: (73.9239287, 0.0),
				75: (74.9215965, 1.0),
				76: (75.922394, 0.0),
				77: (76.9206473, 0.0),
				78: (77.921827, 0.0),
				79: (78.920948, 0.0),
				80: (79.922534, 0.0),
				81: (80.922132, 0.0),
				82: (81.9245, 0.0),
				83: (82.92498, 0.0),
				84: (83.92906, 0.0),
				85: (84.93202, 0.0),
				86: (85.9365, 0.0),
				87: (86.9399, 0.0),
				88: (87.94494, 0.0),
				89: (88.94939, 0.0),
				90: (89.9555, 0.0),
				91: (90.96043, 0.0),
				92: (91.9668, 0.0),
				},
		description=(
				"Metalloid element of group 15. There are three allotropes, yellow, "
				"black, and grey. Reacts with halogens, concentrated oxidizing acids "
				"and hot alkalis. Albertus Magnus is believed to have been the first "
				"to isolate the element in 1250."
				)
		)

Sb = Element(
		51,
		"Sb",
		"Antimony",
		group=15,
		period=5,
		block='p',
		series=5,
		mass=121.76,
		eleneg=2.05,
		eleaffin=1.047401,
		covrad=1.4,
		atmrad=1.53,
		vdwrad=0.0,
		tboil=1860.0,
		tmelt=903.91,
		density=6.69,
		eleconfig="[Kr] 4d10 5s2 5p3",
		oxistates="5, 3*, -3",
		ionenergy=(
				8.6084,
				16.53,
				25.3,
				44.2,
				56.0,
				108.0,
				),
		isotopes={
				103: (102.93969, 0.0),
				104: (103.93647, 0.0),
				105: (104.93149, 0.0),
				106: (105.92879, 0.0),
				107: (106.92415, 0.0),
				108: (107.92216, 0.0),
				109: (108.918132, 0.0),
				110: (109.91675, 0.0),
				111: (110.91316, 0.0),
				112: (111.912398, 0.0),
				113: (112.909372, 0.0),
				114: (113.90927, 0.0),
				115: (114.906598, 0.0),
				116: (115.906794, 0.0),
				117: (116.904836, 0.0),
				118: (117.905529, 0.0),
				119: (118.903942, 0.0),
				120: (119.905072, 0.0),
				121: (120.9038157, 0.5721),
				122: (121.9051737, 0.0),
				123: (122.904214, 0.4279),
				124: (123.9059357, 0.0),
				125: (124.9052538, 0.0),
				126: (125.90725, 0.0),
				127: (126.906924, 0.0),
				128: (127.909169, 0.0),
				129: (128.909148, 0.0),
				130: (129.911656, 0.0),
				131: (130.911982, 0.0),
				132: (131.914467, 0.0),
				133: (132.915252, 0.0),
				134: (133.92038, 0.0),
				135: (134.92517, 0.0),
				136: (135.93035, 0.0),
				137: (136.93531, 0.0),
				138: (137.94079, 0.0),
				139: (138.94598, 0.0),
				},
		description=(
				"Element of group 15. Multiple allotropic forms. The stable form of "
				"antimony is a blue-white metal. Yellow and black antimony are "
				"unstable non-metals. Used in flame-proofing, paints, ceramics, "
				"enamels, and rubber. Attacked by oxidizing acids and halogens. "
				"First reported by Tholden in 1450."
				)
		)

Bi = Element(
		83,
		"Bi",
		"Bismuth",
		group=15,
		period=6,
		block='p',
		series=7,
		mass=208.9804,
		eleneg=2.02,
		eleaffin=0.942363,
		covrad=1.46,
		atmrad=1.63,
		vdwrad=0.0,
		tboil=1837.0,
		tmelt=544.59,
		density=9.8,
		eleconfig="[Xe] 4f14 5d10 6s2 6p3",
		oxistates="5, 3*",
		ionenergy=(
				7.2855,
				16.69,
				25.56,
				45.3,
				56.0,
				88.3,
				),
		isotopes={
				184: (184.00112, 0.0),
				185: (184.99763, 0.0),
				186: (185.9966, 0.0),
				187: (186.993158, 0.0),
				188: (187.99227, 0.0),
				189: (188.9892, 0.0),
				190: (189.9883, 0.0),
				191: (190.985786, 0.0),
				192: (191.98546, 0.0),
				193: (192.98296, 0.0),
				194: (193.98283, 0.0),
				195: (194.980651, 0.0),
				196: (195.980667, 0.0),
				197: (196.978864, 0.0),
				198: (197.97921, 0.0),
				199: (198.977672, 0.0),
				200: (199.978132, 0.0),
				201: (200.977009, 0.0),
				202: (201.977742, 0.0),
				203: (202.976876, 0.0),
				204: (203.977813, 0.0),
				205: (204.977389, 0.0),
				206: (205.978499, 0.0),
				207: (206.9784707, 0.0),
				208: (207.9797422, 0.0),
				209: (208.9803987, 1.0),
				210: (209.9841204, 0.0),
				211: (210.987269, 0.0),
				212: (211.9912857, 0.0),
				213: (212.994385, 0.0),
				214: (213.998712, 0.0),
				215: (215.00177, 0.0),
				216: (216.006306, 0.0),
				217: (217.00947, 0.0),
				218: (218.01432, 0.0),
				},
		description=(
				"White crystalline metal with a pink tinge, belongs to group 15. "
				"Most diamagnetic of all metals and has the lowest thermal "
				"conductivity of all the elements except mercury. Lead-free bismuth "
				"compounds are used in cosmetics and medical procedures. Burns in "
				"the air and produces a blue flame. In 1753, C.G. Junine first "
				"demonstrated that it was different from lead."
				)
		)

Mc = Element(
		115,
		"Mc",
		"Moscovium",
		group=15,
		period=7,
		block='p',
		series=8,
		isotopes={
				287: (287.19119, 0.0),
				288: (288.19249, 0.0),
				289: (289.19272, 0.0),
				290: (290.19414, 0.0),
				291: (291.19438, 0.0),
				}
		)



#!/usr/bin/env python3
#
#  pug_rest.py
"""
Functions for interacting with PubChem PUG_REST API.
"""
#  Copyright (c) 2019-2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on PubChemPy https://github.com/mcs07/PubChemPy/blob/master/LICENSE
#  |  Copyright 2014 Matt Swain <m.swain@me.com>
#  |  Licensed under the MIT License
#  |
#  |  Permission is hereby granted, free of charge, to any person obtaining a copy
#  |  of this software and associated documentation files (the "Software"), to deal
#  |  in the Software without restriction, including without limitation the rights
#  |  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  |  copies of the Software, and to permit persons to whom the Software is
#  |  furnished to do so, subject to the following conditions:
#
#  |  The above copyright notice and this permission notice shall be included in
#  |  all copies or substantial portions of the Software.
#
#  |  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  |  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  |  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  |  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  |  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  |  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  |  THE SOFTWARE.
#

# stdlib
import time
from typing import Dict, Iterable, Iterator, List, Optional, Sequence, Union
from urllib.parse import quote

# 3rd party
import requests

# this package
from chemistry_tools.pubchem import API_BASE
from chemistry_tools.pubchem.enums import PubChemFormats, PubChemNamespace
from chemistry_tools.pubchem.errors import HTTP_ERROR_CODES, PubChemHTTPError
from chemistry_tools.pubchem.utils import _force_sequence_or_csv, _make_base_url

__all__ = ["get_full_json", "async_get", "request", "do_rest_get"]


def do_rest_get(
		namespace: Union[PubChemNamespace, str],
		identifier: Union[str, int, Sequence[Union[str, int]]],
		format_: Union[PubChemFormats, str] = PubChemFormats.JSON,
		domain: Optional[str] = None,
		record_type: str = "2d",
		png_width: int = 300,
		png_height: int = 300,
		) -> requests.Response:
	r"""
	Responsible for performing the actual GET request.

	:param namespace: The type of identifier to look up. Valid values are in :class:`~.PubChemNamespace`.
	:param identifier: Identifiers (e.g. name, CID) for the compounds to look up.
		When using the CID namespace data for multiple compounds can be retrieved at once by
		supplying either a comma-separated string or a list.
	:param format\_: The file format to retrieve the data in.
		Valid values are in :class:`~.PubChemFormats`, plus ``'PNG'``.
	:param domain:
	:param record_type:
	:param png_width:
	:param png_height:
	"""

	# domain = description, synonyms, or property followed by a comma-separated list of desired properties

	if not PubChemNamespace.is_valid_value(namespace):
		raise ValueError(f"'{namespace}' is not a valid value for 'namespace'")

	if not PubChemFormats.is_valid_value(format_):
		raise ValueError(f"'{format_}' is not a valid value for 'format_'")

	parsed_identifier: List[str]

	if namespace == PubChemNamespace.cid:
		parsed_identifier = _force_sequence_or_csv(identifier, "identifier")
	else:
		parsed_identifier = [str(identifier)]

	query_params = {}

	if str(format_).upper() == str(PubChemFormats.PNG):
		query_params["image_size"] = f"{png_width}x{png_height}"

	try:
		r = do_cached_request(namespace, parsed_identifier, format_, domain, record_type, query_params)
	except requests.exceptions.ConnectionError:
		r = do_cached_request(namespace, parsed_identifier, format_, domain, record_type, query_params)

	if r.status_code in HTTP_ERROR_CODES:
		raise PubChemHTTPError(r)

	return r


def do_cached_request(
		namespace: Union[PubChemNamespace, str],
		identifier: Union[Iterable[str], str],
		format_: Union[PubChemFormats, str],
		domain: Optional[str],
		record_type: str,
		query_params: Dict,
		) -> requests.Response:
	r"""
	Responsible for performing cached requests.

	:param namespace: The type of identifier to look up. Valid values are in :class:`~.PubChemNamespace`.
	:param identifier: Identifiers (e.g. name, CID) for the compounds to look up.
		When using the CID namespace data for multiple compounds can be retrieved at once by
		supplying either a comma-separated string or a list.
	:param format\_: The file format to retrieve the data in. Valid values are in :class:`~.PubChemFormats`, plus "PNG"
	:param domain:
	:param record_type:
	:param query_params:
	"""

	if domain:
		r = (_make_base_url(namespace, identifier) / f"{domain}/{format_}").get(params=query_params)
	else:
		query_params["record_type"] = record_type
		r = (_make_base_url(namespace, identifier) / str(format_)).get(params=query_params)

	return r


def get_full_json(cid: Union[str, int]) -> str:
	"""
	Returns the full JSON record for the compound with the given ID.

	:param cid:
	"""

	json_file = (API_BASE / f"_view/data/compound/{cid}/JSON").get()
	return json_file.json()


def async_get(
		identifier,
		namespace: Union[PubChemNamespace, str] = "cid",
		operation=None,
		output="JSON",
		searchtype=None,
		**kwargs
		) -> bytes:
	r"""
	Request wrapper that automatically handles asynchronous requests.

	:param identifier: Identifiers (e.g. name, CID) for the compounds to look up.
		When using the CID namespace data for multiple compounds can be retrieved at once by
		supplying either a comma-separated string or a list.
	:param namespace: The type of identifier to look up. Valid values are in :class:`~.PubChemNamespace`.
	:param operation:
	:param output:
	:param searchtype:
	:param \*\*kwargs: Keyword parameters passed along with the GET request.
	"""

	if (searchtype and searchtype != "xref") or namespace in ["formula"]:
		r = request(identifier, namespace, None, "JSON", searchtype, **kwargs)
		response = r.content
		status = r.json()
		if "Waiting" in status and "ListKey" in status["Waiting"]:
			identifier = status["Waiting"]["ListKey"]
			namespace = "listkey"
			while "Waiting" in status and "ListKey" in status["Waiting"]:
				time.sleep(2)
				r = request(identifier, namespace, operation, "JSON", **kwargs)
				response = r.content
				status = r.json()
			if output != "JSON":
				response = request(identifier, namespace, operation, output, searchtype, **kwargs).content
	else:
		response = request(identifier, namespace, operation, output, searchtype, **kwargs).content

	return response


def request(
		identifier,
		namespace: Union[PubChemNamespace, str] = "cid",
		operation=None,
		output: Union[PubChemFormats, str] = "JSON",
		searchtype=None,
		**kwargs,
		) -> requests.Response:
	r"""
	Construct API request from parameters and return the response.

	Full specification at http://pubchem.ncbi.nlm.nih.gov/pug_rest/PUG_REST.html

	:param identifier: Identifiers (e.g. name, CID) for the compounds to look up.
		When using the CID namespace data for multiple compounds can be retrieved at once by
		supplying either a comma-separated string or a list.
	:param namespace: The type of identifier to look up. Valid values are in :class:`~.PubChemNamespace`.
	:param operation:
	:param output:
	:param searchtype:
	:param \*\*kwargs: Keyword parameters passed along with the GET request.
	"""

	# If identifier is a list, join with commas into string
	if isinstance(identifier, int):
		identifier = str(identifier)

	identifier = _force_sequence_or_csv(identifier, "identifier")
	identifier = ','.join(str(x) for x in identifier)

	# Build API URL
	urlid, params = None, {}

	# use this function when:
	# namespace in ['listkey', 'formula']
	# searchtype == 'xref'

	urlid = quote(identifier.encode("utf8"))

	comps: Iterator[str] = filter(None, ("compound", searchtype, namespace, urlid, operation, output))
	apiurl = API_BASE / '/'.join(comps)

	# Filter None values from kwargs
	for key, val in kwargs.items():
		if val is not None:
			params[key] = val

	# print(f'Request URL: {apiurl}')
	# print(f'Request data: {params}')

	response = apiurl.get(params=params)
	if response.status_code in HTTP_ERROR_CODES:
		raise PubChemHTTPError(response)

	return response



#!/usr/bin/env python3
#
#  compound.py
"""
Parse formulae into a Python object.
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on ChemPy (https://github.com/bjodah/chempy)
#  |  Copyright (c) 2015-2018, Björn Dahlgren
#  |  All rights reserved.
#  |
#  |  Redistribution and use in source and binary forms, with or without modification,
#  |  are permitted provided that the following conditions are met:
#  |
#  |    Redistributions of source code must retain the above copyright notice, this
#  |    list of conditions and the following disclaimer.
#  |
#  |    Redistributions in binary form must reproduce the above copyright notice, this
#  |    list of conditions and the following disclaimer in the documentation and/or
#  |    other materials provided with the distribution.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  |  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  |  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  |  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
#  |  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  |  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  |  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#  |  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  |  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  |  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#  Also based on Pyteomics (https://github.com/levitsky/pyteomics)
#  |  Copyright (c) 2011-2015, Anton Goloborodko & Lev Levitsky
#  |  Licensed under the Apache License, Version 2.0 (the "License");
#  |  you may not use this file except in compliance with the License.
#  |  You may obtain a copy of the License at
#  |
#  |    http://www.apache.org/licenses/LICENSE-2.0
#  |
#  |  Unless required by applicable law or agreed to in writing, software
#  |  distributed under the License is distributed on an "AS IS" BASIS,
#  |  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  |  See the License for the specific language governing permissions and
#  |  limitations under the License.
#  |
#  |  See also:
#  |  Goloborodko, A.A.; Levitsky, L.I.; Ivanov, M.V.; and Gorshkov, M.V. (2013)
#  |  "Pyteomics - a Python Framework for Exploratory Data Analysis and Rapid Software
#  |  Prototyping in Proteomics", Journal of The American Society for Mass Spectrometry,
#  |  24(2), 301–304. DOI: `10.1007/s13361-012-0516-6 <http://dx.doi.org/10.1007/s13361-012-0516-6>`_
#  |
#  |  Levitsky, L.I.; Klein, J.; Ivanov, M.V.; and Gorshkov, M.V. (2018)
#  |  "Pyteomics 4.0: five years of development of a Python proteomics framework",
#  |  Journal of Proteome Research.
#  |  DOI: `10.1021/acs.jproteome.8b00717 <http://dx.doi.org/10.1021/acs.jproteome.8b00717>`_
#

# stdlib
from typing import Dict, Optional

# 3rd party
import quantities  # type: ignore[import]
from domdf_python_tools.bases import Dictable
from domdf_python_tools.doctools import prettify_docstrings

# this package
from chemistry_tools.formulae.formula import Formula
from chemistry_tools.formulae.html import string_to_html
from chemistry_tools.formulae.latex import string_to_latex
from chemistry_tools.formulae.unicode import string_to_unicode

__all__ = ["Compound"]


@prettify_docstrings
class Compound(Dictable):
	"""
	Class representing a chemical compound.

	:param name: The name of the compound
	:param formula: The chemical formula of the compound. If :py:obj:`None` this is generated from the name
	:param data: Free form dictionary.
	:param latex_name:
	:param unicode_name:
	:param html_name:

	.. autosummary-widths:: 4/10

	``data`` could be simple such as ``{'mp': 0, 'bp': 100}`` or considerably more involved,
	e.g.:

	.. code-block:: python

		{
			'diffusion_coefficient': {
				'water': lambda T: 2.1*m**2/s/K*(T - 273.15*K),
			}
		}

	"""

	# 	formula:
	# 		The :class:`~chemistry_tools.formulae.formula.Formula` object representing the compound
	# 	data
	# 		Free form dictionary of additional properties.

	formula: Formula
	latex_name: str
	unicode_name: str
	html_name: str

	def __init__(
			self,
			name: str,
			formula: Optional[Formula] = None,
			data: Optional[Dict] = None,
			latex_name: Optional[str] = None,
			unicode_name: Optional[str] = None,
			html_name: Optional[str] = None,
			):

		super().__init__()

		self.name: str = name

		if formula is None:
			self.formula = Formula.from_string(name)
		else:
			self.formula = formula

		if latex_name:
			self.latex_name = latex_name
		else:
			self.latex_name = string_to_latex(self.formula.hill_formula)

		if unicode_name:
			self.unicode_name = unicode_name
		else:
			self.unicode_name = string_to_unicode(self.formula.hill_formula)

		if html_name:
			self.html_name = html_name
		else:
			self.html_name = string_to_html(self.formula.hill_formula)

		self.data: Optional[Dict] = data or {}

	@property
	def __dict__(self):  # noqa: MAN002
		return dict(
				name=self.name,
				latex_name=self.latex_name,
				unicode_name=self.unicode_name,
				html_name=self.html_name,
				formula=self.formula,
				data=self.data,
				)

	def __eq__(self, other) -> bool:  # noqa: MAN001
		if isinstance(other, str):
			return self.name == other
		else:
			return super().__eq__(other)

	@property
	def charge(self) -> int:
		"""
		The charge of the compound.
		"""

		return self.formula.charge

	@property
	def mass(self) -> float:
		"""
		The mass of the compound.
		"""

		return self.formula.mass

	def molar_mass(self) -> quantities.quantity.Quantity:
		"""
		Returns the molar mass (with units) of the substance.

		:bold-title:`Example:`

		.. code-block:: python

			>>> nh4p = Compound('NH4+')
			>>> import quantities
			>>> nh4p.molar_mass(quantities)
			array(18.0384511...) * g/mol
		"""

		return self.mass * quantities.g / quantities.mol

	def __repr__(self) -> str:
		return f"<{self.__class__.__name__}({self.name}, {self.formula})>"

	def __str__(self) -> str:
		return str(self.name)



#!/usr/bin/env python3
#
#  utils.py
"""
General utility functions.
"""
#
#  Copyright (c) 2019-2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on PubChemPy https://github.com/mcs07/PubChemPy/blob/master/LICENSE
#  |  Copyright 2014 Matt Swain <m.swain@me.com>
#  |  Licensed under the MIT License
#  |
#  |  Permission is hereby granted, free of charge, to any person obtaining a copy
#  |  of this software and associated documentation files (the "Software"), to deal
#  |  in the Software without restriction, including without limitation the rights
#  |  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  |  copies of the Software, and to permit persons to whom the Software is
#  |  furnished to do so, subject to the following conditions:
#
#  |  The above copyright notice and this permission notice shall be included in
#  |  all copies or substantial portions of the Software.
#
#  |  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  |  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  |  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  |  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  |  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  |  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  |  THE SOFTWARE.
#

# stdlib
from typing import Any, Dict, Iterable, List, Sequence, Set, Union

# 3rd party
from apeye.requests_url import RequestsURL

# this package
from chemistry_tools.pubchem import API_BASE
from chemistry_tools.pubchem.enums import PubChemNamespace

__all__ = ["format_string"]


def format_string(stringwithmarkup: Dict[str, Any]) -> str:
	"""
	Convert a PubChem formatted string into an HTML formatted string.

	:param stringwithmarkup:
	"""

	string = list(stringwithmarkup["String"])
	try:
		markup_list = stringwithmarkup["Markup"]
	except KeyError:
		markup_list = []

	for markup in markup_list:
		style = None
		start = markup["Start"]
		end = markup["Length"] + start - 1
		if markup["Type"] == "Italics":
			style = 'i'
		# handle Other formats

		if style is None:
			print(markup)
			continue

		string[start] = f"<{style}>{string[start]}"
		string[end] = f"{string[end]}</{style}>"

	return ''.join(string)


def _force_sequence_or_csv(
		value: Union[str, int, Iterable[Union[str, int]]],
		name: str,
		) -> List[str]:
	"""
	Coerce ``value`` into a list of strings, including splitting comma-separated values.
	If that is not possible a :exc:`ValueError` is raised.

	:param value: The value to coerce to a list of strings.
	:param name: The name of the property. Used for error messages.
	"""

	err_msg = f"Please supply one or more {name}, either as a comma-separated string or a Sequence of strings."

	if not value:
		raise ValueError(err_msg)

	# --

	err_msg = f"'{name}' must be either a comma-separated string or a Sequence of strings"

	parsed_value: List[str]

	if isinstance(value, str):
		parsed_value = value.split(',')

	elif isinstance(value, int):
		parsed_value = [str(value)]

	elif isinstance(value, (Sequence, Set)):

		output: List[str] = []

		for idx, val in enumerate(value):
			if isinstance(val, str):
				output.append(val)
			elif isinstance(val, int):
				output.append(str(val))
			else:
				raise ValueError(err_msg)

		parsed_value = output

	else:
		raise ValueError(err_msg)

	return [val.strip() for val in parsed_value]


def _make_base_url(
		namespace: Union[PubChemNamespace, str],
		identifier: Union[str, int, Iterable[Union[str, int]]],
		) -> RequestsURL:
	"""
	Constructs the base URL for accessing data about a compound.

	:param namespace:
	:param identifier:

	.. versionchanged:: 0.5.0  Now returns a :class:`apeye.requests_url.RequestsURL`.
	"""

	identifier = _force_sequence_or_csv(identifier, "identifier")
	namespace = str(namespace)
	return API_BASE / f"compound/{namespace}/{','.join(identifier)}"



#  !/usr/bin/env python
#
#  __init__.py
r"""
This module provides a wrapper around the PubChem PUG_REST API.

Data for compounds can be accessed using the
:func:`pubchem.lookup.get_compounds <chemistry_tools.pubchem.lookup.get_compounds>` function.

.. _properties table:

The following table lists the various properties that can be obtained from the PubChem API:

.. tabularcolumns:: |l|>{\raggedright\arraybackslash}p{10cm}|

.. collapse:: Click to expand

	.. properties-table::

"""
#
#  Copyright (c) 2019-2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on PubChemPy https://github.com/mcs07/PubChemPy/blob/master/LICENSE
#  |  Copyright 2014 Matt Swain <m.swain@me.com>
#  |  Licensed under the MIT License
#  |
#  |  Permission is hereby granted, free of charge, to any person obtaining a copy
#  |  of this software and associated documentation files (the "Software"), to deal
#  |  in the Software without restriction, including without limitation the rights
#  |  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  |  copies of the Software, and to permit persons to whom the Software is
#  |  furnished to do so, subject to the following conditions:
#  |
#  |  The above copyright notice and this permission notice shall be included in
#  |  all copies or substantial portions of the Software.
#  |
#  |  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  |  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  |  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  |  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  |  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  |  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  |  THE SOFTWARE.
#

# 3rd party
from apeye.requests_url import RequestsURL

# this package
from chemistry_tools.cache import cached_requests

# this package
from . import atom, bond, enums, errors

API_BASE = RequestsURL("https://pubchem.ncbi.nlm.nih.gov/rest/pug")
API_BASE.session = cached_requests

__all__ = [
		"API_BASE",
		"atom",
		"bond",
		"enums",
		"errors",
		]



#!/usr/bin/env python
#
#  cas.py
"""
Functions for working with CAS registry numbers.
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

# stdlib
from typing import Union, no_type_check

__all__ = ["cas_int_to_string", "check_cas_number", "cas_string_to_int"]


def cas_int_to_string(cas_no: int) -> str:
	"""
	Converts an integer CAS registry number to a hyphenated string.

	:param cas_no:
	"""

	cas_no = int(cas_no)

	check_digit = cas_no % 10

	if check_cas_number(cas_no):
		err_msg = (
				f"Invalid CAS Number. Check digit mismatch: expected {check_digit}, "
				f"computed {check_cas_number(cas_no) + check_digit}."
				)
		raise ValueError(err_msg)

	main_value = (cas_no - check_digit) // 10
	block_2 = main_value % 100
	block_1 = (main_value - block_2) // 100

	return f"{block_1}-{block_2}-{check_digit}"


def check_cas_number(cas_no: int) -> int:
	"""
	Checks the CAS registry number to ensure the check digit is valid
	with respect to the rest of the number.

	If the CAS registry number is valid ``0`` is returned.
	If there is a problem the difference between the computed check digit
	and that given as part of the CAS registry number is returned.

	:param cas_no:
	"""  # noqa: D400

	cas_no = abs(int(cas_no))

	check_digit = cas_no % 10
	main_value = (cas_no - check_digit) // 10
	block_2 = main_value % 100
	block_1 = (main_value - block_2) // 100

	last_digit = block_2 % 10

	check_total = last_digit + (((block_2 - last_digit) // 10) * 2)

	for position, digit in enumerate(str(block_1)[::-1]):
		check_total += int(digit) * (position + 3)

	if check_digit == check_total % 10:
		return 0
	else:
		return (check_total % 10) - check_digit


@no_type_check
def cas_string_to_int(cas_no: Union[str, int]) -> int:
	"""
	Converts a hyphenated string CAS registry number to an integer.

	:param cas_no:

	:raises ValueError: If the CAS registry number is invalid.
	"""

	cas_no = str(cas_no)

	block_1, block_2, check_digit = cas_no.split('-')

	check_digit = int(check_digit)

	block_1 = int(block_1) * 1000
	block_2 = int(block_2) * 10

	cas_no = block_1 + block_2 + check_digit

	check_value = check_cas_number(int(cas_no))

	if check_value:
		raise ValueError(
				f"Invalid CAS Number. Check digit mismatch: expected {check_digit}, "
				f"computed {check_value + check_digit}."
				)

	return cas_no



#!/usr/bin/env python3
#
#  cache.py
"""
Cache for HTTP requests.
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

# 3rd party
from apeye import rate_limiter

__all__ = [
		"cache",
		"cache_dir",
		"cached_requests",
		"clear_cache",
		]

#: The cache object.
cache = rate_limiter.HTTPCache("chemistry_tools")

#: Instance of :class:`requests.Session` with a rate limit of 5 requests per second and a 28 day on-disk cache.
cached_requests = cache.session

#: The cache directory
cache_dir = cache.cache_dir


def clear_cache() -> None:
	"""
	Clear the cache.
	"""

	cache.clear()



#!/usr/bin/env python3
#
#  units.py
"""
Functions for handling SI units.
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on ChemPy (https://github.com/bjodah/chempy)
#  |  Copyright (c) 2015-2018, Björn Dahlgren
#  |  All rights reserved.
#  |
#  |  Redistribution and use in source and binary forms, with or without modification,
#  |  are permitted provided that the following conditions are met:
#  |
#  |    Redistributions of source code must retain the above copyright notice, this
#  |    list of conditions and the following disclaimer.
#  |
#  |    Redistributions in binary form must reproduce the above copyright notice, this
#  |    list of conditions and the following disclaimer in the documentation and/or
#  |    other materials provided with the distribution.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  |  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  |  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  |  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
#  |  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  |  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  |  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#  |  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  |  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  |  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

# stdlib
from typing import Tuple, Union

# 3rd party
import numpy
import quantities  # type: ignore[import]
import quantities.markup  # type: ignore[import]

__all__ = [
		"as_latex",
		"compare_equality",
		"allclose",
		"format_string",
		"per100eV",
		"dm",
		"m3",
		"dm3",
		"cm3",
		"nanomolar",
		"molal",
		"micromole",
		"nanomole",
		"kilojoule",
		"kilogray",
		"perMolar_perSecond",
		"umol_per_J",
		"SI_base_registry",
		"dimension_codes",
		"m_math_space",
		"format_si_units",
		]


def as_latex(quant: quantities.quantity.Quantity) -> str:
	r"""
	Returns the LaTeX reperesentation of the unit of a quantity.

	:bold-title:`Example:`

	.. code-block:: python

		>>> print(as_latex(1/quantities.kelvin))
		\mathrm{\frac{1}{K}}
	"""

	# see https://github.com/python-quantities/python-quantities/issues/148
	return quantities.markup.format_units_latex(quant.dimensionality, mult=r"\\cdot").strip('$')


latex_of_unit = as_latex


def compare_equality(
		a: Union[quantities.quantity.Quantity, float],
		b: Union[quantities.quantity.Quantity, float],
		) -> bool:
	"""
	Returns :py:obj:`True` if two arguments are equal.

	Both arguments need to have the same dimensionality.

	:bold-title:`Examples:`

	.. code-block:: python

		>>> km, m = quantities.kilometre, quantities.metre
		>>> compare_equality(3*km, 3)
		False
		>>> compare_equality(3*km, 3000*m)
		True

	:param a:
	:param b:
	"""

	# Workaround for https://github.com/python-quantities/python-quantities/issues/146
	try:
		a + b
	except TypeError:
		# We might be dealing with e.g. None (None + None raises TypeError)
		try:
			len(a)  # type: ignore[arg-type]
		except TypeError:
			# Assumed scalar
			return a == b
		else:
			if len(a) != len(b):  # type: ignore[arg-type]
				return False
			return all(compare_equality(_a, _b) for _a, _b in zip(a, b))  # type: ignore[arg-type]
	except ValueError:
		return False
	else:
		return a == b


def allclose(a, b, rtol=1e-8, atol=None) -> bool:
	"""
	Analogous to :py:func:`numpy.allclose`.

	:param a:
	:param b:
	:param rtol: The relative tolerance.
	:param atol: The absolute tolerance.
	"""

	try:
		d = abs(a - b)
	except Exception:
		try:
			if len(a) == len(b):
				return all(allclose(_a, _b, rtol, atol) for _a, _b in zip(a, b))
			else:
				return False
		except Exception:
			return False
	lim = abs(a) * rtol
	if atol is not None:
		lim += atol

	try:
		len(d)
	except TypeError:
		return d <= lim
	else:
		try:
			len(lim)
		except TypeError:
			return numpy.all([_d <= lim for _d in d])  # type: ignore[return-value]
		else:
			return numpy.all([_d <= _lim for _d, _lim in zip(d, lim)])  # type: ignore[return-value]


# TODO: decide whether to deprecate in favor of "number_to_scientific_latex"?
def format_string(
		value: quantities.quantity.Quantity,
		precision: str = "%.5g",
		tex: bool = False,
		) -> Tuple[str, str]:
	r"""
	Formats a scalar with unit as two strings.

	:bold-title:`Examples:`

	.. code-block:: python

		>>> print(' '.join(format_string(0.42*quantities.mol/decimetre**3)))
		0.42 mol/decimetre**3
		>>> print(' '.join(format_string(2/quantities.s, tex=True)))
		2 \mathrm{\frac{1}{s}}

	:param value: Value with unit
	:param precision:
	:param tex: Whether the string should be formatted for LaTeX.
	"""

	if tex:
		unit_str = as_latex(value)
	else:
		attr = "unicode" if quantities.markup.config.use_unicode else "string"
		unit_str = getattr(value.dimensionality, attr)
	return precision % float(value.magnitude), unit_str


# Additional units to complement quantities
#: Per 100 electronVolts.
per100eV = quantities.UnitQuantity(
		"per_100_eV", 1 / (100 * quantities.eV * quantities.constants.Avogadro_constant), u_symbol="(100eV)**-1"
		)

#: Decimetre
dm = decimetre = quantities.UnitQuantity("decimetre", quantities.m / 10.0, u_symbol="dm")

#: Square metre
m3 = quantities.metre**3

#: Square decimetre
dm3 = decimetre**3

#: Square cenimetre
cm3 = quantities.centimetre**3

#: Nanomolar
nanomolar = quantities.UnitQuantity("nM", 1e-6 * quantities.mole / m3, u_symbol="nM")

#: Molal (moles per kilogram)
molal = quantities.UnitQuantity("molal", quantities.mole / quantities.kg, u_symbol="molal")

#: Micromole
micromole = quantities.UnitQuantity("micromole", quantities.mole / 1e6, u_symbol="μmol")

#: Nanomole
nanomole = quantities.UnitQuantity("nanomole", quantities.mole / 1e9, u_symbol="nmol")

#: Kilojoule
kilojoule = quantities.UnitQuantity("kilojoule", 1e3 * quantities.joule, u_symbol="kJ")

#: Kilogray
kilogray = quantities.UnitQuantity("kilogray", 1e3 * quantities.gray, u_symbol="kGy")

#: Per Molar per second.
perMolar_perSecond = 1 / quantities.molar / quantities.s

#: Micro mole per joule.
umol_per_J = quantities.umol / quantities.joule

SI_base_registry = {
		"length": quantities.metre,
		"mass": quantities.kilogram,
		"time": quantities.second,
		"current": quantities.ampere,
		"temperature": quantities.kelvin,
		"luminous_intensity": quantities.candela,
		"amount": quantities.mole
		}
"""
Mapping of SI measurements to their units.

.. latex:clearpage::
"""

#: Mapping of dimension names to symbols.
dimension_codes = {
		"length": 'L',
		"mass": 'M',
		"time": 'T',
		"current": 'I',
		"temperature": 'Θ',
		"amount": 'N',
		}

m_math_space: str = '\u205f'
"""
A medium mathematical space, ``\u205f``` / ``\\u205f``.

.. versionadded:: 0.4.0
"""


def format_si_units(value: float, *units: str) -> str:
	r"""
	Returns the given value, followed by the given units, and separated by a medium mathematical space.

	:param value:
	:param \*units:

	.. versionadded:: 0.4.0
	"""

	return ''.join([str(value), m_math_space, *units]).rstrip(m_math_space)



#!/usr/bin/env python3
#
#  composition.py
"""
Elemental composition of a :class:`~chemistry_tools.formulae.formula.Formula`.
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

# stdlib
from enum import Enum
from typing import Any, Dict, List

# this package
from chemistry_tools import formulae
from chemistry_tools.elements import ELEMENTS

# this package
from ._parser_core import _make_isotope_string
from .dataarray import DataArray
from .unicode import string_to_unicode
from .utils import split_isotope

__all__ = ["CompositionSort", "Composition"]


class CompositionSort(Enum):
	"""
	Lookup for sorting elemental composition output.
	"""

	symbol = "symbol"
	count = "count"
	rel_mass = "rel_mass"
	mass_fraction = "mass_fraction"
	Symbol = "symbol"
	Count = "count"
	Rel_mass = "rel_mass"
	Mass_fraction = "mass_fraction"
	Rel_Mass = "rel_mass"
	Mass_Fraction = "mass_fraction"

	def __str__(self) -> str:
		return str(self.value)


class Composition(DataArray):
	"""
	Class to represent the elemental composition of a :class:`~chemistry_tools.formulae.formula.Formula`.

	:param formula: A :class:`~chemistry_tools.formulae.formula.Formula` object to create the composition for
	"""

	def __init__(self, formula: "formulae.Formula"):
		data: Dict[str, Dict] = {}

		for isymbol, count in formula.items():
			symbol, isotope = split_isotope(isymbol)
			element = ELEMENTS[symbol]
			mass = element.mass * count
			mass_fraction = mass / formula.mass

			data[isymbol] = dict(
					element=element,
					isotope=isotope,
					count=count,
					rel_mass=mass,
					mass_fraction=mass_fraction,
					)

		super().__init__(formula=formula.hill_formula, data=data)

		self._total_mass: float = formula.mass

	@property
	def total_mass(self) -> float:
		"""
		The total mass of the composition.
		"""

		return self._total_mass

	@property
	def n_elements(self) -> int:
		"""
		The number of elements in the composition.
		"""

		return len(self)

	_as_array_kwargs = {"sort_by", "reverse"}
	_as_table_alignment = ["left", "right", "right", "right"]
	_as_table_float_format = [None, None, ".4f", ".4f"]

	def as_array(
			self,
			sort_by: CompositionSort = CompositionSort.symbol,
			reverse: bool = False,
			) -> List[List[Any]]:
		"""
		Returns the elemental composition as a list of lists.

		:param sort_by: The column to sort by.
		:param reverse: Whether the isotopologues should be sorted in reverse order.
		"""

		output = []
		sorted_data = []

		if sort_by not in CompositionSort:
			raise ValueError(f"Unrecognised value for 'sort_by': {sort_by}")
		elif sort_by == CompositionSort.symbol:
			iterable = sorted(self.keys(), reverse=reverse)
		else:
			iterable = sorted(self.keys(), key=lambda key: self[key][sort_by.value], reverse=reverse)

		for symbol in iterable:
			sorted_data.append(self[symbol])

		for data in sorted_data:
			symbol = _make_isotope_string(data["element"].symbol, data["isotope"])
			rel_mass = f"{data['rel_mass']:0.4f}"
			mass_fraction = f"{data['mass_fraction']:0.4f}"

			output.append([symbol, data["count"], rel_mass, mass_fraction])

		output.insert(0, ["Element", "Count", "Relative Mass", "Mass Fraction"])

		return output

	def __str__(self) -> str:
		table = self.as_table(sort_by=CompositionSort.symbol, reverse=True, tablefmt="fancy_grid")
		return f"\n Elemental Composition for {string_to_unicode(self.formula)}\n{table}"



#!/usr/bin/env python
#
#  property_format.py
"""
Format Physical Properties for Chemicals.
"""
#
#  Copyright (c) 2019-2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

# stdlib
import re
from decimal import Decimal

__all__ = ["degC", "equals", "scientific", "uscg1999", "f2c", "property_format"]

deg_c_re = re.compile(r"(\s*)(deg|DEG)(\s*)(C)")
dec_c_symbol = "°C"
mmath_space = '\u205f'
equals_re = re.compile(r"\s*=\s*")
scientific_regex = re.compile("X10.[0-9]+")
f2c_regex = re.compile(r"\d*\.?\d+ *° *F")


def degC(string: str) -> str:
	"""
	Correctly format a temperature in degrees Celsius.

	:param string:
	"""

	return deg_c_re.sub(f"{mmath_space}{dec_c_symbol}", string)


def equals(string: str) -> str:
	"""
	Correctly format an equals sign in a string.

	:param string:
	"""

	return equals_re.sub(" = ", string)


def scientific(string: str) -> str:
	"""
	Correctly format a string containing scientific notation.

	..TODO: Finish

	:param string:
	"""

	try:
		magnitude = scientific_regex.findall(string)[0].replace("X10", '').replace('-', '−')
	except IndexError:  # no scientific notation to format
		return string

	# print(magnitude)

	return scientific_regex.sub(f"×10<sup>{magnitude}</sup>", string)


def uscg1999(string: str) -> str:
	"""
	Strip references to the USCG dataset from the values.

	:param string:
	"""

	return string.replace("(USCG, 1999)", '')


def f2c(string: str) -> str:
	"""
	Convert temperature strings in degrees Fahrenheit to degrees Celsius.

	:param string:
	"""

	try:
		temperature = f2c_regex.findall(string)[0].replace('F', '').replace('°', '').replace(' ', '')
	except IndexError:
		return string

	# Convert to Celsius and strip trailing 0s and decimal points
	temperature = str((Decimal(temperature) - 32) * (Decimal(5) / Decimal(9)))
	if '.' in temperature:
		temperature = temperature.rstrip('0').rstrip('.')
	return f2c_regex.sub(f"{temperature}°C", string)


def property_format(string: str) -> str:
	"""
	Apply all property formatting functions to the string.

	:param string:
	"""

	return f2c(uscg1999(scientific(degC(equals(string))))).rstrip(' ')



#!/usr/bin/env python3
#
#  tetrels.py
r"""
Group 14: Tetrels, carbon group, crystallogens or adamantogens in the Periodic Table.

.. raw:: latex

	\begin{multicols}{2}

.. data:: C

	:class:`~chemistry_tools.elements.classes.Element` representing Carbon

.. data:: Si

	:class:`~chemistry_tools.elements.classes.Element` representing Silicon

.. data:: Ge

	:class:`~chemistry_tools.elements.classes.Element` representing Germanium

.. data:: Sn

	:class:`~chemistry_tools.elements.classes.Element` representing Tin

.. data:: Pb

	:class:`~chemistry_tools.elements.classes.Element` representing Lead

.. data:: Fl

	:class:`~chemistry_tools.elements.classes.Element` representing Flerovium

.. raw:: latex

	\end{multicols}
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on molmass (https://github.com/cgohlke/molmass)
#  |  Copyright (c) 1990-2020, Christoph Gohlke
#  |  All rights reserved.
#  |  Licensed under the BSD 3-Clause License
#  |  Redistribution and use in source and binary forms, with or without
#  |  modification, are permitted provided that the following conditions are met:
#  |
#  |  1. Redistributions of source code must retain the above copyright notice,
#  |     this list of conditions and the following disclaimer.
#  |
#  |  2. Redistributions in binary form must reproduce the above copyright notice,
#  |     this list of conditions and the following disclaimer in the documentation
#  |     and/or other materials provided with the distribution.
#  |
#  |  3. Neither the name of the copyright holder nor the names of its
#  |     contributors may be used to endorse or promote products derived from
#  |     this software without specific prior written permission.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  |  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  |  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  |  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  |  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  |  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  |  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  |  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  |  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  |  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  |  POSSIBILITY OF SUCH DAMAGE.
#  |
#  Also based on Pyteomics (https://github.com/levitsky/pyteomics)
#  |  Copyright (c) 2011-2015, Anton Goloborodko & Lev Levitsky
#  |  Licensed under the Apache License, Version 2.0 (the "License");
#  |  you may not use this file except in compliance with the License.
#  |  You may obtain a copy of the License at
#  |
#  |    http://www.apache.org/licenses/LICENSE-2.0
#  |
#  |  Unless required by applicable law or agreed to in writing, software
#  |  distributed under the License is distributed on an "AS IS" BASIS,
#  |  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  |  See the License for the specific language governing permissions and
#  |  limitations under the License.
#  |
#  |  See also:
#  |  Goloborodko, A.A.; Levitsky, L.I.; Ivanov, M.V.; and Gorshkov, M.V. (2013)
#  |  "Pyteomics - a Python Framework for Exploratory Data Analysis and Rapid Software
#  |  Prototyping in Proteomics", Journal of The American Society for Mass Spectrometry,
#  |  24(2), 301–304. DOI: `10.1007/s13361-012-0516-6 <http://dx.doi.org/10.1007/s13361-012-0516-6>`_
#  |
#  |  Levitsky, L.I.; Klein, J.; Ivanov, M.V.; and Gorshkov, M.V. (2018)
#  |  "Pyteomics 4.0: five years of development of a Python proteomics framework",
#  |  Journal of Proteome Research.
#  |  DOI: `10.1021/acs.jproteome.8b00717 <http://dx.doi.org/10.1021/acs.jproteome.8b00717>`_
#
#  Isotope data from http://www.nist.gov/pml/data/comp.cfm .

__all__ = ('C', "Si", "Ge", "Sn", "Pb", "Fl")

# this package
from .classes import Element

# Isotope 0 Key:
# mass of the most abundant isotope and 1.0 abundance.

C = Element(
		6,
		'C',
		"Carbon",
		group=14,
		period=2,
		block='p',
		series=1,
		mass=12.01074,
		eleneg=2.55,
		eleaffin=1.262118,
		covrad=0.77,
		atmrad=0.91,
		vdwrad=1.7,
		tboil=5100.0,
		tmelt=3825.0,
		density=3.51,
		eleconfig="[He] 2s2 2p2",
		oxistates="4*, 2, -4*",
		ionenergy=(
				11.2603,
				24.383,
				47.877,
				64.492,
				392.077,
				489.981,
				),
		isotopes={
				8: (8.037675, 0.0),
				9: (9.0310367, 0.0),
				10: (10.0168532, 0.0),
				11: (11.0114336, 0.0),
				12: (12.0, 0.9893),
				13: (13.0033548378, 0.0107),
				14: (14.003241989, 0.0),
				15: (15.0105993, 0.0),
				16: (16.014701, 0.0),
				17: (17.022586, 0.0),
				18: (18.02676, 0.0),
				19: (19.03481, 0.0),
				20: (20.04032, 0.0),
				21: (21.04934, 0.0),
				22: (22.0572, 0.0),
				},
		description=(
				"Carbon is a member of group 14 of the periodic table. It has three "
				"allotropic forms of it, diamonds, graphite and fullerite. Carbon-14 "
				"is commonly used in radioactive dating. Carbon occurs in all "
				"organic life and is the basis of organic chemistry. Carbon has the "
				"interesting chemical property of being able to bond with itself, "
				"and a wide variety of other elements."
				)
		)

Si = Element(
		14,
		"Si",
		"Silicon",
		group=14,
		period=3,
		block='p',
		series=5,
		mass=28.0855,
		eleneg=1.9,
		eleaffin=1.389521,
		covrad=1.11,
		atmrad=1.46,
		vdwrad=2.1,
		tboil=2630.0,
		tmelt=1683.0,
		density=2.33,
		eleconfig="[Ne] 3s2 3p2",
		oxistates="4*, -4",
		ionenergy=(
				8.1517,
				16.345,
				33.492,
				45.141,
				166.77,
				205.05,
				246.52,
				303.17,
				351.1,
				401.43,
				476.06,
				523.5,
				2437.676,
				2673.108,
				),
		isotopes={
				22: (22.03453, 0.0),
				23: (23.02552, 0.0),
				24: (24.011546, 0.0),
				25: (25.004106, 0.0),
				26: (25.99233, 0.0),
				27: (26.98670491, 0.0),
				28: (27.9769265325, 0.92223),
				29: (28.9764947, 0.04685),
				30: (29.97377017, 0.03092),
				31: (30.97536323, 0.0),
				32: (31.97414808, 0.0),
				33: (32.978, 0.0),
				34: (33.978576, 0.0),
				35: (34.98458, 0.0),
				36: (35.9866, 0.0),
				37: (36.99294, 0.0),
				38: (37.99563, 0.0),
				39: (39.00207, 0.0),
				40: (40.00587, 0.0),
				41: (41.01456, 0.0),
				42: (42.01979, 0.0),
				43: (43.02866, 0.0),
				44: (44.03526, 0.0),
				},
		description=(
				"Metalloid element belonging to group 14 of the periodic table. It "
				"is the second most abundant element in the Earth's crust, making up "
				"25.7% of it by weight. Chemically less reactive than carbon. First "
				"identified by Lavoisier in 1787 and first isolated in 1823 by "
				"Berzelius."
				)
		)

Ge = Element(
		32,
		"Ge",
		"Germanium",
		group=14,
		period=4,
		block='p',
		series=5,
		mass=72.63,
		eleneg=2.01,
		eleaffin=1.232712,
		covrad=1.22,
		atmrad=1.52,
		vdwrad=0.0,
		tboil=3107.0,
		tmelt=1211.5,
		density=5.32,
		eleconfig="[Ar] 3d10 4s2 4p2",
		oxistates="4*",
		ionenergy=(7.8994, 15.934, 34.22, 45.71, 93.5),
		isotopes={
				58: (57.99101, 0.0),
				59: (58.98175, 0.0),
				60: (59.97019, 0.0),
				61: (60.96379, 0.0),
				62: (61.95465, 0.0),
				63: (62.94964, 0.0),
				64: (63.94165, 0.0),
				65: (64.93944, 0.0),
				66: (65.93384, 0.0),
				67: (66.932734, 0.0),
				68: (67.928094, 0.0),
				69: (68.9279645, 0.0),
				70: (69.9242474, 0.2038),
				71: (70.924951, 0.0),
				72: (71.9220758, 0.2731),
				73: (72.9234589, 0.0776),
				74: (73.9211778, 0.3672),
				75: (74.9228589, 0.0),
				76: (75.9214026, 0.0783),
				77: (76.9235486, 0.0),
				78: (77.922853, 0.0),
				79: (78.9254, 0.0),
				80: (79.92537, 0.0),
				81: (80.92882, 0.0),
				82: (81.92955, 0.0),
				83: (82.93462, 0.0),
				84: (83.93747, 0.0),
				85: (84.94303, 0.0),
				86: (85.94649, 0.0),
				87: (86.95251, 0.0),
				88: (87.95691, 0.0),
				89: (88.96383, 0.0),
				},
		description=(
				"Lustrous hard metalloid element, belongs to group 14 of the "
				"periodic table. Forms a large number of organometallic compounds. "
				"Predicted by Mendeleev in 1871, it was actually found in 1886 by "
				"Winkler."
				)
		)

Sn = Element(
		50,
		"Sn",
		"Tin",
		group=14,
		period=5,
		block='p',
		series=7,
		mass=118.71,
		eleneg=1.96,
		eleaffin=1.112066,
		covrad=1.41,
		atmrad=1.72,
		vdwrad=2.17,
		tboil=2876.0,
		tmelt=505.12,
		density=7.29,
		eleconfig="[Kr] 4d10 5s2 5p2",
		oxistates="4*, 2*",
		ionenergy=(7.3439, 14.632, 30.502, 40.734, 72.28),
		isotopes={
				99: (98.94933, 0.0),
				100: (99.93904, 0.0),
				101: (100.93606, 0.0),
				102: (101.9303, 0.0),
				103: (102.9281, 0.0),
				104: (103.92314, 0.0),
				105: (104.92135, 0.0),
				106: (105.91688, 0.0),
				107: (106.91564, 0.0),
				108: (107.911925, 0.0),
				109: (108.911283, 0.0),
				110: (109.907843, 0.0),
				111: (110.907734, 0.0),
				112: (111.904818, 0.0097),
				113: (112.905171, 0.0),
				114: (113.902779, 0.0066),
				115: (114.903342, 0.0034),
				116: (115.901741, 0.1454),
				117: (116.902952, 0.0768),
				118: (117.901603, 0.2422),
				119: (118.903308, 0.0859),
				120: (119.9021947, 0.3258),
				121: (120.9042355, 0.0),
				122: (121.903439, 0.0463),
				123: (122.9057208, 0.0),
				124: (123.9052739, 0.0579),
				125: (124.9077841, 0.0),
				126: (125.907653, 0.0),
				127: (126.91036, 0.0),
				128: (127.910537, 0.0),
				129: (128.91348, 0.0),
				130: (129.913967, 0.0),
				131: (130.917, 0.0),
				132: (131.917816, 0.0),
				133: (132.92383, 0.0),
				134: (133.92829, 0.0),
				135: (134.93473, 0.0),
				136: (135.93934, 0.0),
				137: (136.94599, 0.0),
				},
		description=(
				"Silvery malleable metallic element belonging to group 14 of the "
				"periodic table. Twenty-six isotopes are known, five of which are "
				"radioactive. Chemically reactive. Combines directly with chlorine "
				"and oxygen and displaces hydrogen from dilute acids."
				)
		)

Pb = Element(
		82,
		"Pb",
		"Lead",
		group=14,
		period=6,
		block='p',
		series=7,
		mass=207.2,
		eleneg=2.33,
		eleaffin=0.364,
		covrad=1.47,
		atmrad=1.81,
		vdwrad=2.02,
		tboil=2023.0,
		tmelt=600.65,
		density=11.34,
		eleconfig="[Xe] 4f14 5d10 6s2 6p2",
		oxistates="4, 2*",
		ionenergy=(7.4167, 15.032, 31.937, 42.32, 68.8),
		isotopes={
				178: (178.00383, 0.0),
				179: (179.00215, 0.0),
				180: (179.997918, 0.0),
				181: (180.99662, 0.0),
				182: (181.992672, 0.0),
				183: (182.99187, 0.0),
				184: (183.988142, 0.0),
				185: (184.98761, 0.0),
				186: (185.984239, 0.0),
				187: (186.983918, 0.0),
				188: (187.980874, 0.0),
				189: (188.98081, 0.0),
				190: (189.978082, 0.0),
				191: (190.97827, 0.0),
				192: (191.975785, 0.0),
				193: (192.97617, 0.0),
				194: (193.974012, 0.0),
				195: (194.974542, 0.0),
				196: (195.972774, 0.0),
				197: (196.973431, 0.0),
				198: (197.972034, 0.0),
				199: (198.972917, 0.0),
				200: (199.971827, 0.0),
				201: (200.972885, 0.0),
				202: (201.972159, 0.0),
				203: (202.973391, 0.0),
				204: (203.9730436, 0.014),
				205: (204.9744818, 0.0),
				206: (205.9744653, 0.241),
				207: (206.9758969, 0.221),
				208: (207.9766521, 0.524),
				209: (208.9810901, 0.0),
				210: (209.9841885, 0.0),
				211: (210.988737, 0.0),
				212: (211.9918975, 0.0),
				213: (212.996581, 0.0),
				214: (213.9998054, 0.0),
				215: (215.00481, 0.0),
				},
		description=(
				"Heavy dull grey ductile metallic element, belongs to group 14. "
				"Used in building construction, lead-place accumulators, bullets and "
				"shot, and is part of solder, pewter, bearing metals, type metals "
				"and fusible alloys."
				)
		)

Fl = Element(
		114,
		"Fl",
		"Flerovium",
		group=14,
		period=7,
		block='p',
		series=8,
		isotopes={
				285: (285.1837, 0.0),
				286: (286.18386, 0.0),
				287: (287.1856, 0.0),
				288: (288.18569, 0.0),
				289: (289.18728, 0.0),
				}
		)
